# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.fields

import waldur_core.cost_tracking.models


class Migration(migrations.Migration):

    dependencies = [
        ('cost_tracking', '0023_consumptiondetails'),
    ]

    operations = [
        migrations.AddField(
            model_name='priceestimate',
            name='parents',
            field=models.ManyToManyField(help_text='Price estimate parents', related_name='children', to='cost_tracking.PriceEstimate'),
        ),
        migrations.AlterField(
            model_name='defaultpricelistitem',
            name='item_type',
            field=models.CharField(help_text='Type of price list item. Examples: storage, flavor.', max_length=255),
        ),
        migrations.AlterField(
            model_name='defaultpricelistitem',
            name='key',
            field=models.CharField(help_text='Key that corresponds particular consumable. Example: name of flavor.', max_length=255),
        ),
        migrations.AlterField(
            model_name='defaultpricelistitem',
            name='metadata',
            field=waldur_core.core.fields.JSONField(blank=True, default={}, help_text='Details of the item, that corresponds price list item. Example: details of flavor.'),
        ),
        migrations.AlterField(
            model_name='priceestimate',
            name='consumed',
            field=models.FloatField(default=0, help_text='Price for resource until now.'),
        ),
        migrations.AlterField(
            model_name='priceestimate',
            name='details',
            field=waldur_core.core.fields.JSONField(default={}, help_text='Saved scope details. Field is populated on scope deletion.'),
        ),
        migrations.AlterField(
            model_name='priceestimate',
            name='limit',
            field=models.FloatField(default=-1, help_text='How many funds object can consume in current month."-1" means no limit.'),
        ),
        migrations.AlterField(
            model_name='priceestimate',
            name='total',
            field=models.FloatField(default=0, help_text='Predicted price for scope for current month.'),
        ),
        migrations.AlterUniqueTogether(
            name='defaultpricelistitem',
            unique_together=set([('key', 'item_type', 'resource_content_type')]),
        ),
        migrations.AlterUniqueTogether(
            name='priceestimate',
            unique_together=set([('content_type', 'object_id', 'month', 'year')]),
        ),
        migrations.RemoveField(
            model_name='priceestimate',
            name='is_manually_input',
        ),
        migrations.RemoveField(
            model_name='priceestimate',
            name='is_visible',
        ),
        migrations.RemoveField(
            model_name='priceestimate',
            name='leafs',
        ),
        migrations.RemoveField(
            model_name='priceestimate',
            name='scope_customer',
        ),
        migrations.AlterField(
            model_name='consumptiondetails',
            name='configuration',
            field=waldur_core.cost_tracking.models.ConsumableItemsField(default={}, help_text='Current resource configuration.'),
        ),
        migrations.AlterField(
            model_name='consumptiondetails',
            name='consumed_before_update',
            field=waldur_core.cost_tracking.models.ConsumableItemsField(default={}, help_text='How many consumables were used by resource before last update.'),
        ),
    ]



from __future__ import absolute_import

import codecs
import csv

import six
from six import StringIO


class UTF8Recoder:
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8
    """

    def __init__(self, f, encoding):
        self.reader = codecs.getreader(encoding)(f)

    def __iter__(self):
        return self

    def next(self):
        return next(self.reader).encode("utf-8")


class UnicodeDictReader:
    """
    A CSV reader which will iterate over lines in the CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, encoding="utf-8", **kwargs):
        f = UTF8Recoder(f, encoding)
        self.reader = csv.reader(f, **kwargs)
        self.header = next(self.reader)

    def next(self):
        row = next(self.reader)
        vals = [six.text_type(s, "utf-8") for s in row]
        return dict((self.header[x], vals[x]) for x in range(len(self.header)))

    def __iter__(self):
        return self


class UnicodeDictWriter:
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, fieldnames, encoding="utf-8", **kwargs):
        # Redirect output to a queue
        self.fieldnames = fieldnames
        self.queue = StringIO()
        self.writer = csv.writer(self.queue, **kwargs)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writeheader(self):
        self.writer.writerow(self.fieldnames)

    def writerow(self, row):
        self.writer.writerow([six.text_type(row[x]).encode("utf-8") for x in self.fieldnames])
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)



from __future__ import unicode_literals

from django.urls import reverse
import factory

from waldur_core.logging import models
# Dependency from `structure` application exists only in tests
from waldur_core.logging.loggers import get_valid_events
from waldur_core.structure.tests import factories as structure_factories


class EventFactory(object):
    """
    Event factory that provides default data for events.

    Created event fields can be accessible via .fields attribute of created event.
    """

    def __init__(self, **kwargs):
        self.create(**kwargs)
        self.save()

    def create(self, **kwargs):
        """
        Creates event fields values.

        If field is in kwargs - value from kwargs will be used for this field,
        otherwise - default value will be used for field.
        """
        self.fields = {
            '@timestamp': '2015-04-19T16:25:45.376+04:00',
            '@version': 1,
            'customer_abbreviation': 'TCAN',
            'customer_contact_details': 'test details',
            'customer_name': 'Test cusomter',
            'customer_uuid': 'test_customer_uuid',
            'event_type': 'test_event_type',
            'host': 'example.com',
            'importance': 'high',
            'importance_code': 30,
            'levelname': 'WARNING',
            'logger': 'waldur_core.test',
            'message': 'Test message',
            'project_name': 'test_project',
            'project_uuid': 'test_project_uuid',
            'tags': ['_jsonparsefailure'],
            'type': 'gcloud-event',
            'user_uuid': 'test_user_uuid',
        }
        for key, value in kwargs.items():
            self.fields[key] = value

    @classmethod
    def get_list_url(cls):
        return 'http://testserver' + reverse('event-list')


class AlertFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.Alert

    message = factory.Sequence(lambda i: 'message#%s' % i)
    alert_type = factory.Iterator(['first_alert', 'second_alert', 'third_alert', 'fourth_alert'])
    severity = factory.Iterator([
        models.Alert.SeverityChoices.DEBUG, models.Alert.SeverityChoices.INFO,
        models.Alert.SeverityChoices.WARNING, models.Alert.SeverityChoices.ERROR])
    context = {'test': 'test'}
    scope = factory.SubFactory(structure_factories.CustomerFactory)

    @classmethod
    def get_list_url(cls):
        return 'http://testserver' + reverse('alert-list')

    @classmethod
    def get_url(cls, alert=None, action=None):
        if alert is None:
            alert = AlertFactory()
        url = 'http://testserver' + reverse('alert-detail', kwargs={'uuid': alert.uuid})
        return url if action is None else url + action + '/'


class WebHookFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.WebHook

    event_types = get_valid_events()[:3]
    destination_url = 'http://example.com/'

    @classmethod
    def get_list_url(cls):
        return 'http://testserver' + reverse('webhook-list')

    @classmethod
    def get_url(cls, hook=None):
        if hook is None:
            hook = WebHookFactory()
        return 'http://testserver' + reverse('webhook-detail', kwargs={'uuid': hook.uuid})


class PushHookFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.PushHook

    event_types = get_valid_events()[:3]
    token = 'VALID_TOKEN'
    type = models.PushHook.Type.ANDROID

    @classmethod
    def get_list_url(cls):
        return 'http://testserver' + reverse('pushhook-list')

    @classmethod
    def get_url(cls, hook=None):
        if hook is None:
            hook = PushHookFactory()
        return 'http://testserver' + reverse('pushhook-detail', kwargs={'uuid': hook.uuid})



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


def drop_customers_for_shared_service_settings(apps, schema_editor):
    ServiceSettings = apps.get_model('structure', 'ServiceSettings')
    for settings in ServiceSettings.objects.all():
        if settings.shared and settings.customer is not None:
            settings.customer = None
            settings.save()


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0045_project_services_certifications'),
    ]

    operations = [
        migrations.RunPython(drop_customers_for_shared_service_settings),
    ]



from datetime import timedelta
from ddt import ddt, data
from django.utils import timezone
from rest_framework import test, status
from reversion.models import Version

from waldur_core.core import utils as core_utils
from waldur_core.quotas.tests import factories
from waldur_core.structure import models as structure_models
from waldur_core.structure.tests import (factories as structure_factories,
                                         fixtures as structure_fixtures)


@ddt
class QuotaUpdateTest(test.APITransactionTestCase):
    def setUp(self):
        super(QuotaUpdateTest, self).setUp()
        self.fixture = structure_fixtures.ServiceFixture()
        self.quota_name = structure_models.Customer.Quotas.nc_project_count
        self.quota = self.fixture.customer.quotas.get(name=self.quota_name)
        self.quota.usage = 10
        self.quota.save()
        self.url = factories.QuotaFactory.get_url(self.quota)

    def test_staff_can_set_quota_limit(self):
        self.client.force_authenticate(self.fixture.staff)
        response = self.client.put(self.url, {'limit': self.quota.usage + 1})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['limit'], self.quota.usage + 1)

    @data('global_support', 'owner')
    def test_other_users_can_not_set_quota_limit(self, user):
        self.client.force_authenticate(getattr(self.fixture, user))
        response = self.client.put(self.url, {'limit': self.quota.usage + 1})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_quota_limit_should_not_be_less_than_usage(self):
        self.client.force_authenticate(self.fixture.staff)
        response = self.client.put(self.url, {'limit': self.quota.usage - 1})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)


class QuotaHistoryTest(test.APITransactionTestCase):

    def setUp(self):
        self.customer = structure_factories.CustomerFactory()
        self.owner = structure_factories.UserFactory(username='owner')
        self.customer.add_user(self.owner, structure_models.CustomerRole.OWNER)

        self.quota = factories.QuotaFactory(scope=self.customer)
        self.url = factories.QuotaFactory.get_url(self.quota, 'history')
        # Hook for test: lets say that revision was created one hour ago
        version = Version.objects.get_for_object(self.quota).filter(revision__date_created__lte=timezone.now()).first()
        version.revision.date_created = timezone.now() - timedelta(hours=1)
        version.revision.save()

    def test_old_version_of_quota_is_available(self):
        old_usage = self.quota.usage
        self.quota.usage = self.quota.usage + 1
        self.quota.save()
        history_timestamp = core_utils.datetime_to_timestamp(timezone.now() - timedelta(minutes=30))

        self.client.force_authenticate(self.owner)
        response = self.client.get(self.url, data={'point': history_timestamp})

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data[0]['point'], history_timestamp)
        self.assertEqual(response.data[0]['object']['usage'], old_usage)

    def test_endpoint_does_not_return_object_if_date(self):
        history_timestamp = core_utils.datetime_to_timestamp(timezone.now() - timedelta(hours=2))

        self.client.force_authenticate(self.owner)
        response = self.client.get(self.url, data={'point': history_timestamp})

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertNotIn('object', response.data[0])

    def test_timeinterval_is_splitted_equal_parts(self):
        start_timestamp = 1436094000
        end_timestamp = 1436096000

        self.client.force_authenticate(self.owner)
        response = self.client.get(self.url, data={'points_count': 3, 'start': start_timestamp, 'end': end_timestamp})

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data[0]['point'], start_timestamp)
        self.assertEqual(response.data[1]['point'], start_timestamp + (end_timestamp - start_timestamp) / 2)
        self.assertEqual(response.data[2]['point'], end_timestamp)


# TODO: add CRUD tests for quota endpoint.



import copy
import json
from datetime import datetime

from django.contrib.admin.sites import AdminSite
from django.test import TestCase

from waldur_core.structure import admin as structure_admin
from waldur_core.structure import models as structure_models
from waldur_core.structure.tests import fixtures, factories
from waldur_core.structure.tests.serializers import ServiceSerializer
from waldur_core.structure.utils import get_all_services_field_info, FieldInfo


class MockRequest:
    pass


class MockSuperUser:
    def has_perm(self, perm):
        return True


request = MockRequest()
request.user = MockSuperUser()


class override_serializer(object):
    def __init__(self, field_info):
        self.field_info = field_info
        self.required = copy.copy(ServiceSerializer.Meta.required_fields)

        if ServiceSerializer.SERVICE_ACCOUNT_FIELDS is not NotImplemented:
            self.fields = copy.copy(ServiceSerializer.SERVICE_ACCOUNT_FIELDS)
        else:
            self.fields = NotImplemented

        if ServiceSerializer.SERVICE_ACCOUNT_EXTRA_FIELDS is not NotImplemented:
            self.extra_fields = copy.copy(ServiceSerializer.SERVICE_ACCOUNT_EXTRA_FIELDS)
        else:
            self.extra_fields = NotImplemented

    def __enter__(self):
        ServiceSerializer.Meta.required_fields = self.field_info.fields_required
        ServiceSerializer.SERVICE_ACCOUNT_FIELDS = {
            field: ''
            for field in self.field_info.fields
        }
        ServiceSerializer.SERVICE_ACCOUNT_EXTRA_FIELDS = {
            field: ''
            for field in self.field_info.extra_fields_required
        }
        return ServiceSerializer

    def __exit__(self, *args):
        ServiceSerializer.Meta.required_fields = self.required
        ServiceSerializer.SERVICE_ACCOUNT_FIELDS = self.fields
        ServiceSerializer.SERVICE_ACCOUNT_EXTRA_FIELDS = self.extra_fields


class ServiceSettingAdminTest(TestCase):
    def setUp(self):
        super(ServiceSettingAdminTest, self).setUp()
        get_all_services_field_info.cache_clear()

    def test_if_required_field_value_is_provided_form_is_valid(self):
        fields = FieldInfo(
            fields_required=['backend_url'],
            fields=['backend_url'],
            extra_fields_required=[]
        )

        data = self.get_valid_data(backend_url='http://test.net')
        self.assert_form_valid(fields, data)

    def test_if_required_field_value_is_not_provided_form_is_invalid(self):
        fields = FieldInfo(
            fields_required=['backend_url'],
            fields=['backend_url'],
            extra_fields_required=[]
        )

        data = self.get_valid_data()
        self.assert_form_invalid(fields, data)

    def test_if_required_extra_field_value_is_provided_form_is_valid(self):
        fields = FieldInfo(
            fields_required=['tenant'],
            fields=[],
            extra_fields_required=['tenant']
        )
        data = self.get_valid_data(options=json.dumps({'tenant': 1}))
        self.assert_form_valid(fields, data)

    def test_if_required_extra_field_value_is_not_provided_form_is_invalid(self):
        fields = FieldInfo(
            fields_required=['tenant'],
            fields=[],
            extra_fields_required=['tenant']
        )
        data = self.get_valid_data()
        self.assert_form_invalid(fields, data)

    def test_if_options_is_not_valid_json_form_is_invalid(self):
        fields = FieldInfo(
            fields_required=['tenant'],
            fields=[],
            extra_fields_required=['tenant']
        )
        data = self.get_valid_data(options='INVALID')
        self.assert_form_invalid(fields, data)

    def get_valid_data(self, **kwargs):
        data = {
            'type': 'Test',
            'name': 'test',
            'state': 1,
            'username': 'test',
            'password': 'xxx',
            'options': json.dumps({}),
        }
        data.update(kwargs)
        return data

    def form_is_valid(self, fields, data):
        with override_serializer(fields):
            site = AdminSite()
            model_admin = structure_admin.PrivateServiceSettingsAdmin(
                structure_models.PrivateServiceSettings, site)
            form = model_admin.get_form(request)(data)
            return form.is_valid()

    def assert_form_valid(self, fields, data):
        self.assertTrue(self.form_is_valid(fields, data))

    def assert_form_invalid(self, fields, data):
        self.assertFalse(self.form_is_valid(fields, data))


class ProjectAdminTest(TestCase):
    def setUp(self):
        super(ProjectAdminTest, self).setUp()
        self.fixture = fixtures.ProjectFixture()
        self.project = self.fixture.project
        self.created_by = factories.UserFactory()

    def change_project(self, **kwargs):
        site = AdminSite()
        model_admin = structure_admin.ProjectAdmin(structure_models.Project, site)

        request = MockRequest()
        request.user = self.created_by

        form_for_data = model_admin.get_form(request)(instance=self.project)
        post_data = form_for_data.initial
        post_data.update(kwargs)

        form = model_admin.get_form(request)(instance=self.project, data=post_data)
        form.save()

        self.project.refresh_from_db()
        return self.project

    def test_new_users_are_added(self):
        # Arrange
        user1 = factories.UserFactory()
        user2 = factories.UserFactory()

        # Act
        project = self.change_project(support_users=[user1.pk, user2.pk])

        # Asset
        self.assertTrue(project.has_user(user1, structure_models.ProjectRole.SUPPORT))
        self.assertTrue(project.has_user(user2, structure_models.ProjectRole.SUPPORT))

    def test_old_users_are_deleted_and_existing_are_preserved(self):
        # Arrange
        user1 = factories.UserFactory()
        user2 = factories.UserFactory()
        self.project.add_user(user1, structure_models.ProjectRole.SUPPORT)

        # Act
        project = self.change_project(support_users=[user2.pk])

        # Asset
        self.assertFalse(project.has_user(user1, structure_models.ProjectRole.SUPPORT))
        self.assertTrue(project.has_user(user2, structure_models.ProjectRole.SUPPORT))

    def test_user_may_have_only_one_role_in_the_same_project(self):
        # Arrange
        user1 = factories.UserFactory()
        user2 = factories.UserFactory()

        # Act
        with self.assertRaises(ValueError):
            self.change_project(support_users=[user1.pk, user2.pk], managers=[user1.pk, user2.pk])


class CustomerAdminTest(TestCase):
    def setUp(self):
        super(CustomerAdminTest, self).setUp()
        self.fixture = fixtures.CustomerFixture()
        self.customer = self.fixture.customer
        self.created_by = factories.UserFactory()

    def change_customer(self, **kwargs):
        site = AdminSite()
        model_admin = structure_admin.CustomerAdmin(structure_models.Customer, site)

        request = MockRequest()
        request.user = self.created_by

        form_for_data = model_admin.get_form(request)(instance=self.customer)
        post_data = form_for_data.initial

        dt_now = datetime.now()
        dt_date = dt_now.strftime("%Y-%m-%d")
        dt_time = dt_now.strftime("%H:%M:%S")

        post_data.update(dict(
            accounting_start_date_0=dt_date,
            accounting_start_date_1=dt_time
        ))

        post_data.update(kwargs)

        form = model_admin.get_form(request)(instance=self.customer, data=post_data)
        form.save()

        self.customer.refresh_from_db()
        return self.customer

    def test_new_users_are_added(self):
        # Arrange
        user1 = factories.UserFactory()
        user2 = factories.UserFactory()

        # Act
        customer = self.change_customer(support_users=[user1.pk, user2.pk])

        # Asset
        self.assertTrue(customer.has_user(user1, structure_models.CustomerRole.SUPPORT))
        self.assertTrue(customer.has_user(user2, structure_models.CustomerRole.SUPPORT))

    def test_old_users_are_deleted_and_existing_are_preserved(self):
        # Arrange
        user1 = factories.UserFactory()
        user2 = factories.UserFactory()
        self.customer.add_user(user1, structure_models.CustomerRole.SUPPORT)

        # Act
        customer = self.change_customer(support_users=[user2.pk])

        # Asset
        self.assertFalse(customer.has_user(user1, structure_models.CustomerRole.SUPPORT))
        self.assertTrue(customer.has_user(user2, structure_models.CustomerRole.SUPPORT))

    def test_user_may_have_only_one_role_in_the_same_customer(self):
        # Arrange
        user1 = factories.UserFactory()
        user2 = factories.UserFactory()

        # Act
        with self.assertRaises(ValueError):
            self.change_customer(support_users=[user1.pk, user2.pk], owners=[user1.pk, user2.pk])



from __future__ import unicode_literals

from django.apps import AppConfig
from django.db.models import signals


class EventsConfig(AppConfig):
    name = 'waldur_core.logging'
    verbose_name = 'Logging'

    def ready(self):
        from waldur_core.logging import handlers, utils

        for index, model in enumerate(utils.get_loggable_models()):
            signals.post_delete.connect(
                handlers.remove_related_alerts,
                sender=model,
                dispatch_uid='waldur_core.logging.handlers.remove_{}_{}_related_alerts'.format(model.__name__, index),
            )



from django.contrib import admin
from django.contrib.auth import get_user_model
from django.test import TestCase
from django.urls import reverse


User = get_user_model()


class TestAdminEndpoints(TestCase):

    def setUp(self):
        user, _ = User.objects.get_or_create(username='username', is_staff=True)
        self.client.force_login(user)
        self.admin_site_name = admin.site.name

    def _reverse_url(self, path):
        return reverse('%s:%s' % (self.admin_site_name, path))

    def test_app_list_ulrs_can_be_queried(self):
        app_list_urls = dict()
        for model in admin.site._registry:
            app_list_url = reverse('%s:%s' % (self.admin_site_name, 'app_list'), args=(model._meta.app_label,))
            app_list_urls.update({model._meta.app_label: app_list_url})

        for url in app_list_urls.values():
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)

    def test_base_admin_site_urls_can_be_queried(self):
        pages = ['index', 'login', 'logout', 'password_change', 'password_change_done', 'jsi18n']
        for name in pages:
            url = self._reverse_url(name)
            response = self.client.get(url)
            self.assertIn(response.status_code, [200, 302])

    def test_changelist_urls_can_be_queried(self):
        for model in admin.site._registry:
            url = self._reverse_url('%s_%s_changelist' % (model._meta.app_label, model._meta.model_name))
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)

    def test_add_urls_can_be_queried(self):
        for model in admin.site._registry:
            model_fullname = '%s_%s' % (model._meta.app_label, model._meta.model_name)
            url = self._reverse_url('%s_add' % model_fullname)
            response = self.client.get(url)
            self.assertIn(response.status_code, [200, 403])



from rest_framework import test

from waldur_core.structure import models as structure_models
from waldur_core.structure.tests import factories as structure_factories

from . import factories
from .. import CostTrackingRegister, models


class BaseCostTrackingTest(test.APITransactionTestCase):

    def setUp(self):
        self.users = {
            'staff': structure_factories.UserFactory(username='staff', is_staff=True),
            'owner': structure_factories.UserFactory(username='owner'),
            'administrator': structure_factories.UserFactory(username='administrator'),
            'manager': structure_factories.UserFactory(username='manager'),
        }

        self.customer = structure_factories.CustomerFactory()
        self.customer.add_user(self.users['owner'], structure_models.CustomerRole.OWNER)
        self.project = structure_factories.ProjectFactory(customer=self.customer)
        self.project.add_user(self.users['administrator'], structure_models.ProjectRole.ADMINISTRATOR)
        self.project.add_user(self.users['manager'], structure_models.ProjectRole.MANAGER)

        self.service = structure_factories.TestServiceFactory(customer=self.customer)
        self.service_project_link = structure_factories.TestServiceProjectLinkFactory(
            project=self.project, service=self.service)

        CostTrackingRegister.register_strategy(factories.TestNewInstanceCostTrackingStrategy)
        models.DefaultPriceListItem.init_from_registered_resources()



import os

from django.conf import settings
from django.core.management.base import BaseCommand

from waldur_core.core.docs import ApiDocs


class Command(BaseCommand):
    help = "Generate RST docs for DRF API"

    def add_arguments(self, parser):
        parser.add_argument(
            '--store', '-s', action='store', dest='path',
            default='docs/drfapi', help='Where to store docs.'
        )
        parser.add_argument('args', metavar='app_label', nargs='*', help='Application label.')

    def handle(self, *app_labels, **options):
        path = options.get('path')
        path = path if path.startswith('/') else os.path.join(settings.BASE_DIR, path)

        if not os.path.isdir(path):
            os.makedirs(path)
        else:
            for f in os.listdir(path):
                if f.endswith(".rst"):
                    os.remove(os.path.join(path, f))

        self.stdout.write(self.style.MIGRATE_HEADING('Gather endpoints info'))
        docs = ApiDocs(apps=app_labels)
        self.stdout.write(self.style.MIGRATE_HEADING('Write RST docs'))
        docs.generate(path)



from django.test import TestCase

from waldur_core.quotas import models
from waldur_core.structure import models as structure_models
from waldur_core.structure.tests import factories as structure_factories


class GlobalQuotasHandlersTestCase(TestCase):

    def test_project_global_quota_increased_after_project_creation(self):
        quota = models.Quota.objects.get(name=structure_models.Project.GLOBAL_COUNT_QUOTA_NAME)

        structure_factories.ProjectFactory()

        reread_quota = models.Quota.objects.get(pk=quota.pk)
        self.assertEqual(reread_quota.usage, quota.usage + 1)

    def test_project_global_quota_decreased_after_project_deletion(self):
        project = structure_factories.ProjectFactory()
        quota = models.Quota.objects.get(name=structure_models.Project.GLOBAL_COUNT_QUOTA_NAME)

        project.delete()

        reread_quota = models.Quota.objects.get(pk=quota.pk)
        self.assertEqual(reread_quota.usage, quota.usage - 1)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import django.utils.timezone
import model_utils.fields
from django.conf import settings
from django.db import migrations, models

import waldur_core.core.fields


class Migration(migrations.Migration):
    replaces = [('logging', '0001_squashed_0003_emailhook_webhook'),
                ('logging', '0001_initial'),
                ('logging', '0002_alert_acknowledged'),
                ('logging', '0002_index_alert_type'),
                ('logging', '0003_emailhook_webhook'),
                ('logging', '0003_add_alert_unique_together_constraint'),
                ('logging', '0004_index_alert_type'),
                ('logging', '0005_add_alert_unique_together_constraint'),
                ('logging', '0006_pushhook_systemnotification'), ('logging', '0007_pushhook_unique_key'),
                ('logging', '0008_pushhook_token'), ('logging', '0009_add_device_metadata'),
                ('logging', '0010_add_event_groups')]

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Alert',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('alert_type', models.CharField(db_index=True, max_length=50)),
                ('message', models.CharField(max_length=255)),
                ('severity', models.SmallIntegerField(choices=[(10, 'Debug'), (20, 'Info'), (30, 'Warning'), (40, 'Error')])),
                ('closed', models.DateTimeField(blank=True, null=True)),
                ('is_closed', models.CharField(blank=True, max_length=32)),
                ('acknowledged', models.BooleanField(default=False)),
                ('context', waldur_core.core.fields.JSONField(blank=True)),
                ('object_id', models.PositiveIntegerField(null=True)),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='EmailHook',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('event_types', waldur_core.core.fields.JSONField(verbose_name='List of event types')),
                ('event_groups', waldur_core.core.fields.JSONField(default=[], verbose_name='List of event groups')),
                ('is_active', models.BooleanField(default=True)),
                ('last_published', models.DateTimeField(default=django.utils.timezone.now)),
                ('email', models.EmailField(max_length=75)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PushHook',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('event_types', waldur_core.core.fields.JSONField(verbose_name='List of event types')),
                ('event_groups', waldur_core.core.fields.JSONField(default=[], verbose_name='List of event groups')),
                ('is_active', models.BooleanField(default=True)),
                ('last_published', models.DateTimeField(default=django.utils.timezone.now)),
                ('type', models.SmallIntegerField(choices=[(1, 'iOS'), (2, 'Android')])),
                ('device_id', models.CharField(max_length=255, null=True, unique=True)),
                ('device_manufacturer', models.CharField(blank=True, max_length=255, null=True)),
                ('device_model', models.CharField(blank=True, max_length=255, null=True)),
                ('token', models.CharField(max_length=255, null=True, unique=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='SystemNotification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('event_types', waldur_core.core.fields.JSONField(verbose_name='List of event types')),
                ('event_groups', waldur_core.core.fields.JSONField(default=[], verbose_name='List of event groups')),
                ('hook_content_type', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='WebHook',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('event_types', waldur_core.core.fields.JSONField(verbose_name='List of event types')),
                ('event_groups', waldur_core.core.fields.JSONField(default=[], verbose_name='List of event groups')),
                ('is_active', models.BooleanField(default=True)),
                ('last_published', models.DateTimeField(default=django.utils.timezone.now)),
                ('destination_url', models.URLField()),
                ('content_type', models.SmallIntegerField(choices=[(1, 'json'), (2, 'form')], default=1)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AlterUniqueTogether(
            name='pushhook',
            unique_together=set([('user', 'device_id', 'type')]),
        ),
        migrations.AlterUniqueTogether(
            name='alert',
            unique_together=set([('content_type', 'object_id', 'alert_type', 'is_closed')]),
        ),
    ]



from django.conf.urls import include, url

from waldur_core.core.routers import SortedDefaultRouter as DefaultRouter
from waldur_core.server.urls import urlpatterns

from . import views


def register_in(router):
    router.register(r'test', views.TestServiceViewSet, base_name='test')
    router.register(r'test-service-project-link', views.TestServiceProjectLinkViewSet, base_name='test-spl')
    router.register(r'test-new-instances', views.TestNewInstanceViewSet, base_name='test-new-instances')


router = DefaultRouter()
register_in(router)

urlpatterns += [
    url(r'^api/', include(router.urls)),
]



from django.core.management import call_command
from django.test import TestCase

from waldur_core.structure.tests import factories as structure_factories


class RecalculateCommandTest(TestCase):

    def test_counter_quota_recalculation(self):
        customer = structure_factories.CustomerFactory()
        structure_factories.ProjectFactory(customer=customer)

        customer.quotas.filter(name='nc_project_count').update(usage=10)

        call_command('recalculatequotas')
        self.assertEqual(customer.quotas.get(name='nc_project_count').usage, 1)

    def test_aggregator_quota_recalculation(self):
        customer = structure_factories.CustomerFactory()
        structure_factories.ProjectFactory(customer=customer)

        customer.quotas.filter(name='nc_resource_count').update(usage=10)

        call_command('recalculatequotas')
        self.assertEqual(customer.quotas.get(name='nc_resource_count').usage, 0)



import collections

from django.contrib.contenttypes.models import ContentType
from django.core.management.base import BaseCommand
from django.db.models import Q
import six
from six.moves import input

from waldur_core.cost_tracking import CostTrackingRegister
from waldur_core.cost_tracking.models import PriceEstimate
from waldur_core.structure import models as structure_models


# XXX: This command is not working properly now.
#      We need to decide is it necessary with new structure.
class Command(BaseCommand):
    """
    This management command removes following price estimates:

    1) Price estimates with invalid content type.
       For example, resource plugin has been uninstalled,
       but its price estimates remained.

    2) Price estimates without valid scope and without details.
       For example, resource has been deleted, but its
       price estimate doesn't contain updated details,
       or its price estimates should be deleted.

    3) Price estimates for invalid month.
       Price estimates for each month should contain price estimate
       for at least one resource, one service, and one project.
       Otherwise it is considered invalid.
    """
    help = 'Delete invalid price estimates'

    def add_arguments(self, parser):
        parser.add_argument('--assume-yes', dest='assume_yes', action='store_true')
        parser.set_defaults(assume_yes=False)

    def handle(self, assume_yes, **options):
        self.assume_yes = assume_yes
        self.delete_price_estimates_for_invalid_content_types()
        self.delete_price_estimates_without_scope_and_details()
        self.delete_price_estimates_for_invalid_month()

    def confirm(self):
        if self.assume_yes:
            return True
        confirm = input('Enter [y] to continue: ')
        return confirm.strip().lower() == 'y'

    def delete_price_estimates_for_invalid_month(self):
        invalid_estimates = self.get_all_estimates_wihout_scope_in_month()
        count = invalid_estimates.count()
        if count:
            self.stdout.write('{} price estimates without scope in month would be deleted.'.
                              format(count))
            if self.confirm():
                invalid_estimates.delete()

    def get_all_estimates_wihout_scope_in_month(self):
        invalid_estimates = []
        for customer in structure_models.Customer.objects.all():
            customer_estimates = self.get_estimates_without_scope_in_month(customer)
            invalid_estimates.extend(customer_estimates)
        ids = [estimate.pk for estimate in invalid_estimates]
        return PriceEstimate.objects.filter(pk__in=ids)

    def get_estimated_models(self):
        return (
            structure_models.Customer,
            structure_models.ServiceSettings,
            structure_models.Service,
            structure_models.ServiceProjectLink,
            structure_models.Project,
        ) + tuple(CostTrackingRegister.registered_resources.keys())

    def get_estimates_without_scope_in_month(self, customer):
        """
        It is expected that valid row for each month contains at least one
        price estimate for customer, service setting, service,
        service project link, project and resource.
        Otherwise all price estimates in the row should be deleted.
        """
        estimates = self.get_price_estimates_for_customer(customer)
        if not estimates:
            return []

        tables = {model: collections.defaultdict(list)
                  for model in self.get_estimated_models()}

        dates = set()
        for estimate in estimates:
            date = (estimate.year, estimate.month)
            dates.add(date)

            cls = estimate.content_type.model_class()
            for model, table in tables.items():
                if issubclass(cls, model):
                    table[date].append(estimate)
                    break

        invalid_estimates = []
        for date in dates:
            if any(map(lambda table: len(table[date]) == 0, tables.values())):
                for table in tables.values():
                    invalid_estimates.extend(table[date])
        print(invalid_estimates)
        return invalid_estimates

    def get_price_estimates_for_customer(self, customer):
        descendants_estimates = []
        customer_descendants = customer.get_descendants()
        for descendant in customer_descendants:
            descendants_estimates += list(PriceEstimate.objects.filter(scope=descendant))
        customer_estimates = PriceEstimate.objects.filter(scope=customer)
        return list(customer_estimates) + descendants_estimates

    def delete_price_estimates_without_scope_and_details(self):
        invalid_estimates = self.get_invalid_price_estimates()
        count = invalid_estimates.count()
        if count:
            self.stdout.write('{} price estimates without scope and details would be deleted.'.
                              format(count))
            if self.confirm():
                invalid_estimates.delete()

    def get_invalid_price_estimates(self):
        query = Q(details='', object_id=None)
        for model in PriceEstimate.get_estimated_models():
            content_type = ContentType.objects.get_for_model(model)
            ids = set(model.objects.all().values_list('id', flat=True))
            if ids:
                query |= Q(content_type=content_type, object_id__in=ids)
        return PriceEstimate.objects.all().exclude(query)

    def delete_price_estimates_for_invalid_content_types(self):
        content_types = self.get_invalid_content_types()
        content_types_list = ', '.join(map(six.text_type, content_types))

        query = Q(content_type__in=content_types) | Q(content_type__isnull=True)
        invalid_estimates = PriceEstimate.objects.all().filter(query).filter()
        count = invalid_estimates.count()

        if count:
            self.stdout.write('{} price estimates for invalid content types would be deleted: {}'.
                              format(count, content_types_list))
            if self.confirm():
                invalid_estimates.delete()

    def get_invalid_content_types(self):
        valid = [
            ContentType.objects.get_for_model(model)
            for model in PriceEstimate.get_estimated_models()
        ]
        invalid = set(
            PriceEstimate.objects.all()
            .exclude(content_type__in=valid)
            .distinct()
            .values_list('content_type_id', flat=True)
        )

        return ContentType.objects.all().filter(id__in=invalid)



from django.test import TransactionTestCase
from freezegun import freeze_time

from waldur_core.cost_tracking import models, ConsumableItem
from waldur_core.cost_tracking.tests import factories


class ConsumptionDetailsManagerTest(TransactionTestCase):

    def test_create_method_get_configuration_from_previous_month_details(self):
        with freeze_time("2016-08-01"):
            configuration = {ConsumableItem('storage', '1 MB'): 10240, ConsumableItem('ram', '1 MB'): 2048}
            price_estimate = factories.PriceEstimateFactory(year=2016, month=8)
            consumption_details = factories.ConsumptionDetailsFactory(price_estimate=price_estimate)
            consumption_details.update_configuration(configuration)

        with freeze_time("2016-09-01"):
            next_price_estimate = models.PriceEstimate.objects.create(
                scope=price_estimate.scope,
                month=price_estimate.month + 1,
                year=price_estimate.year,
            )
            next_consumption_details = models.ConsumptionDetails.objects.create(price_estimate=next_price_estimate)
        self.assertDictEqual(next_consumption_details.configuration, configuration)



import datetime

from django.contrib.contenttypes.models import ContentType
from django.test import TransactionTestCase
from django.utils import timezone
from freezegun import freeze_time

from waldur_core.cost_tracking import CostTrackingRegister, models, ConsumableItem, tasks
from waldur_core.cost_tracking.tests import factories
from waldur_core.structure.tests import factories as structure_factories
from waldur_core.structure.tests.models import TestNewInstance


class ResourceUpdateTest(TransactionTestCase):

    def setUp(self):
        CostTrackingRegister.register_strategy(factories.TestNewInstanceCostTrackingStrategy)
        resource_content_type = ContentType.objects.get_for_model(TestNewInstance)
        self.price_list_item = models.DefaultPriceListItem.objects.create(
            item_type='storage', key='1 MB', value=0.5, resource_content_type=resource_content_type)

    @freeze_time('2016-08-08 11:00:00', tick=True)  # freeze time to avoid bugs in the end of a month.
    def test_consumption_details_of_resource_is_keeped_up_to_date(self):
        today = timezone.now()
        configuration = dict(ram=2048, disk=20 * 1024, cores=2)
        resource = structure_factories.TestNewInstanceFactory(
            state=TestNewInstance.States.OK, runtime_state='online', **configuration)

        price_estimate = models.PriceEstimate.objects.get(scope=resource, month=today.month, year=today.year)
        consumption_details = price_estimate.consumption_details
        expected = {
            ConsumableItem('ram', '1 MB'): 2048,
            ConsumableItem('storage', '1 MB'): 20 * 1024,
            ConsumableItem('cores', '1 core'): 2,
            ConsumableItem('quotas', 'test_quota'): 0,
        }
        self.assertDictEqual(consumption_details.configuration, expected)

        resource.ram = 1024
        resource.save()
        consumption_details.refresh_from_db()
        self.assertEqual(consumption_details.configuration[ConsumableItem('ram', '1 MB')], resource.ram)

        resource.runtime_state = 'offline'
        resource.save()
        # test resource uses only storage and quota when it is offline
        expected = {
            ConsumableItem('storage', '1 MB'): 20 * 1024,
            ConsumableItem('quotas', 'test_quota'): 0
        }
        consumption_details.refresh_from_db()
        self.assertDictEqual(consumption_details.configuration, expected)

        resource.flavor_name = 'small'
        resource.save()
        consumption_details.refresh_from_db()
        self.assertEqual(consumption_details.configuration[ConsumableItem('flavor', 'small')], 1)

    def test_price_estimate_of_resource_is_keeped_up_to_date(self):
        start_time = datetime.datetime(2016, 8, 8, 11, 0)
        with freeze_time(start_time):
            today = timezone.now()
            old_disk = 20 * 1024
            resource = structure_factories.TestNewInstanceFactory(
                state=TestNewInstance.States.OK, runtime_state='online', disk=old_disk)
            price_estimate = models.PriceEstimate.objects.get(scope=resource, month=today.month, year=today.year)
            # after resource creation price estimate should be calculate for whole month
            month_end = datetime.datetime(2016, 8, 31, 23, 59, 59)
            expected = (
                int((month_end - start_time).total_seconds() / 60) * self.price_list_item.minute_rate * old_disk)
            self.assertAlmostEqual(price_estimate.total, expected)

        # after some time resource disk was updated - resource price should be recalculated
        change_time = datetime.datetime(2016, 8, 9, 13, 0)
        with freeze_time(change_time):
            new_disk = 40 * 1024
            resource.disk = new_disk
            resource.save()

            price_estimate.refresh_from_db()
            expected = (
                int((change_time - start_time).total_seconds() / 60) * self.price_list_item.minute_rate * old_disk +
                int((month_end - change_time).total_seconds() / 60) * self.price_list_item.minute_rate * new_disk)
            self.assertAlmostEqual(price_estimate.total, expected)

    def test_price_estimate_of_resource_ancestors_is_keeped_up_to_date(self):
        """ On resource configuration tests handlers should update resource ancestors estimates """
        start_time = datetime.datetime(2016, 8, 8, 11, 0)
        with freeze_time(start_time):
            today = timezone.now()
            old_disk = 20 * 1024
            resource = structure_factories.TestNewInstanceFactory(
                state=TestNewInstance.States.OK, runtime_state='online', disk=old_disk)
            ancestors = [resource.service_project_link, resource.service_project_link.service,
                         resource.service_project_link.project, resource.service_project_link.project.customer]
            # after resource creation price for it ancestors should be calculate for whole month
            month_end = datetime.datetime(2016, 8, 31, 23, 59, 59)
            expected = (
                int((month_end - start_time).total_seconds() / 60) * self.price_list_item.minute_rate * old_disk)
            for ancestor in ancestors:
                ancestor_estimate = models.PriceEstimate.objects.get(scope=ancestor, month=today.month, year=today.year)
                self.assertAlmostEqual(ancestor_estimate.total, expected)

        # after some time resource disk was updated - resource ancestors price should be recalculated
        change_time = datetime.datetime(2016, 8, 9, 13, 0)
        with freeze_time(change_time):
            new_disk = 40 * 1024
            resource.disk = new_disk
            resource.save()

            expected = (
                int((change_time - start_time).total_seconds() / 60) * self.price_list_item.minute_rate * old_disk +
                int((month_end - change_time).total_seconds() / 60) * self.price_list_item.minute_rate * new_disk)
            for ancestor in ancestors:
                ancestor_estimate = models.PriceEstimate.objects.get(scope=ancestor, month=today.month, year=today.year)
                self.assertAlmostEqual(ancestor_estimate.total, expected)

    def test_historical_estimates_are_initialized(self):
        creation_time = timezone.make_aware(datetime.datetime(2016, 7, 15, 11, 0))
        import_time = timezone.make_aware(datetime.datetime(2016, 9, 2, 10, 0))
        with freeze_time(import_time):
            resource = structure_factories.TestNewInstanceFactory(
                state=TestNewInstance.States.OK, runtime_state='online', disk=20 * 1024, created=creation_time)
            ancestors = [resource.service_project_link, resource.service_project_link.service,
                         resource.service_project_link.project, resource.service_project_link.project.customer]

        # signal should init estimates for resource and its ancestors for previous months
        for scope in [resource] + ancestors:
            self.assertTrue(models.PriceEstimate.objects.filter(scope=scope, month=7, year=2016))
            self.assertTrue(models.PriceEstimate.objects.filter(scope=scope, month=8, year=2016))

        # Check price estimates total calculation for month #7
        month_end = timezone.make_aware(datetime.datetime(2016, 7, 31, 23, 59, 59))
        work_minutes = int((month_end - creation_time).total_seconds() / 60)
        expected = work_minutes * self.price_list_item.minute_rate * resource.disk
        for scope in [resource] + ancestors:
            estimate = models.PriceEstimate.objects.get(scope=scope, month=7, year=2016)
            self.assertEqual(estimate.total, expected)

        # Check price estimates total calculation for month #8
        month_start = timezone.make_aware(datetime.datetime(2016, 8, 1))
        month_end = timezone.make_aware(datetime.datetime(2016, 8, 31, 23, 59, 59))
        work_minutes = int((month_end - month_start).total_seconds() / 60)
        expected = work_minutes * self.price_list_item.minute_rate * resource.disk
        for scope in [resource] + ancestors:
            estimate = models.PriceEstimate.objects.get(scope=scope, month=8, year=2016)
            self.assertAlmostEqual(estimate.total, expected)


class ResourceQuotaUpdateTest(TransactionTestCase):

    def setUp(self):
        CostTrackingRegister.register_strategy(factories.TestNewInstanceCostTrackingStrategy)

    @freeze_time('2016-08-08 11:00:00', tick=True)  # freeze time to avoid bugs in the end of a month.
    def test_consumption_details_of_resource_is_keeped_up_to_date_on_quota_change(self):
        today = timezone.now()
        resource = structure_factories.TestNewInstanceFactory()
        quota_item = ConsumableItem('quotas', 'test_quota')

        price_estimate = models.PriceEstimate.objects.get(scope=resource, month=today.month, year=today.year)
        consumption_details = price_estimate.consumption_details
        self.assertEqual(consumption_details.configuration[quota_item], 0)

        resource.set_quota_usage(TestNewInstance.Quotas.test_quota, 5)

        consumption_details.refresh_from_db()
        self.assertEqual(consumption_details.configuration[quota_item], 5)


class ScopeDeleteTest(TransactionTestCase):

    def setUp(self):
        resource_content_type = ContentType.objects.get_for_model(TestNewInstance)
        self.price_list_item = models.DefaultPriceListItem.objects.create(
            item_type='storage', key='1 MB', resource_content_type=resource_content_type, value=0.1)
        CostTrackingRegister.register_strategy(factories.TestNewInstanceCostTrackingStrategy)
        self.start_time = datetime.datetime(2016, 8, 8, 11, 0)
        with freeze_time(self.start_time):
            self.resource = structure_factories.TestNewInstanceFactory(disk=20 * 1024)
        self.spl = self.resource.service_project_link
        self.project = self.spl.project
        self.customer = self.project.customer
        self.service = self.spl.service

    @freeze_time('2016-08-08 13:00:00')
    def test_all_estimates_are_deleted_on_customer_deletion(self):
        self.resource.delete()
        self.project.delete()
        self.customer.delete()

        for scope in (self.spl, self.resource, self.service, self.project, self.customer):
            self.assertFalse(models.PriceEstimate.objects.filter(scope=scope).exists(),
                             'Unexpected price estimate exists for %s %s' % (scope.__class__.__name__, scope))

    @freeze_time('2016-08-08 13:00:00')
    def test_estimate_is_recalculated_on_resource_deletion(self):
        self.resource.delete()

        expected = 2 * self.price_list_item.value * self.resource.disk  # resource has been working for 2 hours
        for scope in (self.spl, self.service, self.project, self.customer):
            self.assertEqual(models.PriceEstimate.objects.get_current(scope).total, expected)

    @freeze_time('2016-08-08 13:00:00')
    def test_estimate_populate_details_on_scope_deletion(self):
        scopes = (self.resource, self.project)
        for scope in scopes:
            estimate = models.PriceEstimate.objects.get_current(scope)
            scope.delete()
            estimate.refresh_from_db()

            self.assertEqual(estimate.details['name'], scope.name)

        estimate = models.PriceEstimate.objects.get_current(self.service)
        self.service.delete()
        estimate.refresh_from_db()
        self.assertEqual(estimate.details['name'], str(self.service))

    # set time to next month to make sure that estimates for previous months are deleted too.
    @freeze_time('2016-09-01 12:00:00')
    def test_estimated_is_removed_on_resource_unlink(self):
        tasks.recalculate_estimate()  # recalculate to create new estimates in new month.

        self.resource.unlink()
        self.resource.delete()

        self.assertFalse(models.PriceEstimate.objects.filter(scope=self.resource).exists())
        details_names = [estimate.details.get('name') for estimate in models.PriceEstimate.objects.all()]
        self.assertNotIn(self.resource.name, details_names)

        for scope in (self.spl, self.service, self.project, self.customer):
            for price_estimate in models.PriceEstimate.objects.filter(scope=scope):
                self.assertEqual(price_estimate.total, 0)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('quotas', '0003_index_quota_name'),
    ]

    operations = [
        migrations.AddField(
            model_name='quota',
            name='threshold',
            field=models.FloatField(default=0, validators=[django.core.validators.MinValueValidator(0)]),
        ),
    ]



from __future__ import unicode_literals

import uuid

from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
import django_filters
from django_filters.constants import EMPTY_VALUES
from rest_framework import filters

from waldur_core.core import filters as core_filters
from waldur_core.cost_tracking import models, serializers
from waldur_core.structure import models as structure_models, SupportedServices


class PriceEstimateScopeFilterBackend(core_filters.GenericKeyFilterBackend):

    def get_related_models(self):
        return models.PriceEstimate.get_estimated_models()

    def get_field_name(self):
        return 'scope'


class PriceEstimateDateFilterBackend(filters.BaseFilterBackend):

    def filter_queryset(self, request, queryset, view):
        if 'date' in request.query_params:
            date_serializer = serializers.PriceEstimateDateFilterSerializer(
                data={'date_list': request.query_params.getlist('date')})
            date_serializer.is_valid(raise_exception=True)
            query = Q()
            for year, month in date_serializer.validated_data['date_list']:
                query |= Q(year=year, month=month)
            queryset = queryset.filter(query)

        # Filter by date range
        date_range_serializer = serializers.PriceEstimateDateRangeFilterSerializer(data=request.query_params)
        date_range_serializer.is_valid(raise_exception=True)
        if 'start' in date_range_serializer.validated_data:
            year, month = date_range_serializer.validated_data['start']
            queryset = queryset.filter(Q(year__gt=year) | Q(year=year, month__gte=month))
        if 'end' in date_range_serializer.validated_data:
            year, month = date_range_serializer.validated_data['end']
            queryset = queryset.filter(Q(year__lt=year) | Q(year=year, month__lte=month))

        return queryset


class PriceEstimateCustomerFilterBackend(filters.BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        if 'customer' not in request.query_params:
            return queryset

        customer_uuid = request.query_params['customer']
        try:
            uuid.UUID(customer_uuid)
        except ValueError:
            return queryset.none()

        try:
            customer = structure_models.Customer.objects.get(uuid=customer_uuid)
        except structure_models.Customer.DoesNotExist:
            return queryset.none()

        ids = []
        for estimate in models.PriceEstimate.objects.filter(scope=customer):
            ids.append(estimate.pk)
            for child in estimate.collect_children():
                ids.append(child.pk)
        return queryset.filter(pk__in=ids)


class PriceListItemServiceFilterBackend(core_filters.GenericKeyFilterBackend):

    def get_related_models(self):
        return structure_models.Service.get_all_models()

    def get_field_name(self):
        return 'service'


class ResourceTypeFilter(django_filters.CharFilter):

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs
        resource_models = SupportedServices.get_resource_models()
        try:
            model = resource_models[value]
            ct = ContentType.objects.get_for_model(model)
            return super(ResourceTypeFilter, self).filter(qs, ct)
        except (ContentType.DoesNotExist, KeyError):
            return qs.none()


class DefaultPriceListItemFilter(django_filters.FilterSet):
    resource_content_type = core_filters.ContentTypeFilter()
    resource_type = ResourceTypeFilter(name='resource_content_type')

    class Meta:
        model = models.DefaultPriceListItem
        fields = [
            'key',
            'item_type',
            'resource_content_type',
            'resource_type',
        ]



from rest_framework import test
from six.moves import mock

from waldur_core.core import filters as core_filters


class TestUrlFilter(test.APITransactionTestCase):
    def setUp(self):
        from waldur_core.structure.tests.factories import CustomerFactory

        self.customer = CustomerFactory()
        self.url = CustomerFactory.get_url(self.customer)

        self.customer_filter = core_filters.URLFilter(
            view_name='customer-detail',
            name='customer__uuid'
        )

    def test_filter_checks_that_url_matches_view(self):
        qs = mock.Mock()
        self.customer_filter.filter(qs, self.url)
        qs.filter.assert_called_once_with(customer__uuid__exact=self.customer.uuid.hex)



from collections import defaultdict

from rest_framework import serializers

from .models import ResourceItem, ResourceSla
from .utils import get_period, to_list


class ResourceSlaStateTransitionSerializer(serializers.Serializer):
    timestamp = serializers.IntegerField()
    state = serializers.SerializerMethodField()

    def get_state(self, obj):
        return obj.state and 'U' or 'D'


class MonitoringSerializerMixin(serializers.Serializer):
    sla = serializers.SerializerMethodField()
    monitoring_items = serializers.SerializerMethodField()

    class Meta:
        fields = ('sla', 'monitoring_items')

    def get_sla(self, resource):
        if not hasattr(self, 'sla_map_cache'):
            self.sla_map_cache = {}
            request = self.context['request']

            items = ResourceSla.objects.filter(scope__in=to_list(self.instance))
            items = items.filter(period=get_period(request))
            for item in items:
                self.sla_map_cache[item.object_id] = dict(
                    value=item.value,
                    agreed_value=item.agreed_value,
                    period=item.period
                )

        return self.sla_map_cache.get(resource.id)

    def get_monitoring_items(self, resource):
        if not hasattr(self, 'monitoring_items_map'):
            self.monitoring_items_map = {}
            items = ResourceItem.objects.filter(scope__in=to_list(self.instance))

            self.monitoring_items_map = defaultdict(dict)
            for item in items:
                self.monitoring_items_map[item.object_id][item.name] = item.value

        return self.monitoring_items_map.get(resource.id)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ddt import data, ddt
from django.test import TransactionTestCase
from django.urls import reverse
from freezegun import freeze_time
from mock_django import mock_signal_receiver
from rest_framework import status, test

from waldur_core.core.tests.helpers import override_waldur_core_settings
from waldur_core.quotas.tests import factories as quota_factories
from waldur_core.structure import signals
from waldur_core.structure.models import Customer, CustomerRole, ProjectRole
from waldur_core.structure.tests import factories, fixtures


class UrlResolverMixin(object):
    def _get_customer_url(self, customer):
        return 'http://testserver' + reverse('customer-detail', kwargs={'uuid': customer.uuid})

    def _get_project_url(self, project):
        return 'http://testserver' + reverse('project-detail', kwargs={'uuid': project.uuid})

    def _get_user_url(self, user):
        return 'http://testserver' + reverse('user-detail', kwargs={'uuid': user.uuid})


@freeze_time('2017-11-01 00:00:00')
class CustomerTest(TransactionTestCase):
    def setUp(self):
        self.customer = factories.CustomerFactory()
        self.user = factories.UserFactory()
        self.created_by = factories.UserFactory()

    def test_add_user_returns_created_if_grant_didnt_exist_before(self):
        _, created = self.customer.add_user(self.user, CustomerRole.OWNER)

        self.assertTrue(created, 'Customer permission should have been reported as created')

    def test_add_user_returns_not_created_if_grant_existed_before(self):
        self.customer.add_user(self.user, CustomerRole.OWNER)
        _, created = self.customer.add_user(self.user, CustomerRole.OWNER)

        self.assertFalse(created, 'Customer permission should have been reported as not created')

    def test_add_user_returns_membership(self):
        membership, _ = self.customer.add_user(self.user, CustomerRole.OWNER)

        self.assertEqual(membership.user, self.user)
        self.assertEqual(membership.customer, self.customer)

    def test_add_user_returns_same_membership_for_consequent_calls_with_same_arguments(self):
        membership1, _ = self.customer.add_user(self.user, CustomerRole.OWNER)
        membership2, _ = self.customer.add_user(self.user, CustomerRole.OWNER)

        self.assertEqual(membership1, membership2)

    def test_add_user_emits_structure_role_granted_if_grant_didnt_exist_before(self):
        with mock_signal_receiver(signals.structure_role_granted) as receiver:
            self.customer.add_user(self.user, CustomerRole.OWNER, self.created_by)

        receiver.assert_called_once_with(
            structure=self.customer,
            user=self.user,
            role=CustomerRole.OWNER,

            sender=Customer,
            signal=signals.structure_role_granted,
            created_by=self.created_by
        )

    def test_add_user_doesnt_emit_structure_role_granted_if_grant_existed_before(self):
        self.customer.add_user(self.user, CustomerRole.OWNER)

        with mock_signal_receiver(signals.structure_role_granted) as receiver:
            self.customer.add_user(self.user, CustomerRole.OWNER)

        self.assertFalse(receiver.called, 'structure_role_granted should not be emitted')

    def test_remove_user_emits_structure_role_revoked_for_each_role_user_had_in_customer(self):
        self.customer.add_user(self.user, CustomerRole.OWNER)

        with mock_signal_receiver(signals.structure_role_revoked) as receiver:
            self.customer.remove_user(self.user, removed_by=self.created_by)

        receiver.assert_called_once_with(
            structure=self.customer,
            user=self.user,
            role=CustomerRole.OWNER,
            removed_by=self.created_by,

            sender=Customer,
            signal=signals.structure_role_revoked,
        )

    def test_remove_user_emits_structure_role_revoked_if_grant_existed_before(self):
        self.customer.add_user(self.user, CustomerRole.OWNER)

        with mock_signal_receiver(signals.structure_role_revoked) as receiver:
            self.customer.remove_user(self.user, CustomerRole.OWNER, self.created_by)

        receiver.assert_called_once_with(
            structure=self.customer,
            user=self.user,
            role=CustomerRole.OWNER,
            removed_by=self.created_by,

            sender=Customer,
            signal=signals.structure_role_revoked,
        )

    def test_remove_user_doesnt_emit_structure_role_revoked_if_grant_didnt_exist_before(self):
        with mock_signal_receiver(signals.structure_role_revoked) as receiver:
            self.customer.remove_user(self.user, CustomerRole.OWNER)

        self.assertFalse(receiver.called, 'structure_role_remove should not be emitted')


class CustomerRoleTest(TransactionTestCase):
    def setUp(self):
        self.customer = factories.CustomerFactory()

    def test_get_owners_returns_empty_list(self):
        self.assertEqual(0, self.customer.get_owners().count())


@ddt
class CustomerApiPermissionTest(UrlResolverMixin, test.APITransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ProjectFixture()

    # List filtration tests
    @data('staff', 'global_support', 'owner', 'customer_support', 'admin', 'manager', 'project_support')
    def test_user_can_list_customers(self, user):
        self.client.force_authenticate(user=getattr(self.fixture, user))

        self._check_user_list_access_customers(self.fixture.customer, 'assertIn')

    @data('user', 'admin', 'manager', 'project_support')
    def test_user_cannot_list_other_customer(self, user):
        customer = factories.CustomerFactory()
        self.client.force_authenticate(user=getattr(self.fixture, user))
        self._check_customer_in_list(customer, False)

    # Nested objects filtration tests
    @data('admin', 'manager', 'project_support')
    def test_user_can_see_project_he_has_a_role_in_within_customer(self, user):
        self.client.force_authenticate(user=getattr(self.fixture, user))

        response = self.client.get(self._get_customer_url(self.fixture.customer))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        project_urls = set([project['url'] for project in response.data['projects']])
        self.assertIn(
            self._get_project_url(self.fixture.project), project_urls,
            'User should see project',
        )

    @data('admin', 'manager', 'project_support')
    def test_user_cannot_see_project_he_has_no_role_in_within_customer(self, user):
        self.client.force_authenticate(user=getattr(self.fixture, user))

        non_seen_project = factories.ProjectFactory(customer=self.fixture.customer)

        response = self.client.get(self._get_customer_url(self.fixture.customer))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        project_urls = set([project['url'] for project in response.data['projects']])
        self.assertNotIn(
            self._get_project_url(non_seen_project), project_urls,
            'User should not see project',
        )

    # Direct instance access tests
    @data(('owner', 'owners'), ('customer_support', 'support_users'))
    def test_user_can_see_its_owner_membership_in_a_service_he_is_owner_of(self, user_data):
        user, response_field = user_data
        self.client.force_authenticate(user=getattr(self.fixture, user))
        response = self.client.get(self._get_customer_url(self.fixture.customer))
        users = set(c['url'] for c in response.data[response_field])

        user_url = self._get_user_url(getattr(self.fixture, user))
        self.assertEqual([user_url], list(users))

    @data('staff', 'global_support')
    def test_user_can_access_all_customers_if_he_is_staff(self, user):
        self.client.force_authenticate(user=getattr(self.fixture, user))

        self._check_user_direct_access_customer(self.fixture.customer, status.HTTP_200_OK)

        customer = factories.CustomerFactory()
        self._check_user_direct_access_customer(customer, status.HTTP_200_OK)

    # Helper methods
    def _check_user_list_access_customers(self, customer, test_function):
        response = self.client.get(reverse('customer-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        urls = set([instance['url'] for instance in response.data])
        url = self._get_customer_url(customer)
        getattr(self, test_function)(url, urls)

    def _check_customer_in_list(self, customer, positive=True):
        response = self.client.get(reverse('customer-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        urls = set([instance['url'] for instance in response.data])
        customer_url = self._get_customer_url(customer)
        if positive:
            self.assertIn(customer_url, urls)
        else:
            self.assertNotIn(customer_url, urls)

    def _check_user_direct_access_customer(self, customer, status_code):
        response = self.client.get(self._get_customer_url(customer))
        self.assertEqual(response.status_code, status_code)


@ddt
class CustomerApiManipulationTest(UrlResolverMixin, test.APITransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ProjectFixture()

    # Deletion tests
    @data('owner', 'admin', 'manager', 'global_support', 'customer_support', 'project_support')
    def test_user_cannot_delete_customer(self, user):
        self.client.force_authenticate(user=getattr(self.fixture, user))

        response = self.client.delete(self._get_customer_url(self.fixture.customer))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_user_cannot_delete_customer_with_associated_projects_if_he_is_staff(self):
        self.client.force_authenticate(user=self.fixture.staff)

        factories.ProjectFactory(customer=self.fixture.customer)
        response = self.client.delete(self._get_customer_url(self.fixture.customer))

        self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
        self.assertDictContainsSubset({'detail': 'Cannot delete organization with existing projects'},
                                      response.data)

    # Creation tests
    @data('user', 'global_support')
    def test_user_can_not_create_customer_if_he_is_not_staff(self, user):
        self.client.force_authenticate(user=getattr(self.fixture, user))

        response = self.client.post(factories.CustomerFactory.get_list_url(), self._get_valid_payload())
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_waldur_core_settings(CREATE_DEFAULT_PROJECT_ON_ORGANIZATION_CREATION=True)
    def test_default_project_is_created_if_configured(self):
        self.client.force_authenticate(user=self.fixture.staff)

        response = self.client.post(factories.CustomerFactory.get_list_url(), self._get_valid_payload())

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        customer = Customer.objects.get(uuid=response.data['uuid'])
        self.assertEqual(customer.projects.count(), 1)
        self.assertEqual(customer.projects.first().name, 'First project')
        self.assertEqual(customer.projects.first().description, 'First project we have created for you')

    @override_waldur_core_settings(CREATE_DEFAULT_PROJECT_ON_ORGANIZATION_CREATION=False)
    def test_default_project_is_not_created_if_configured(self):
        self.client.force_authenticate(user=self.fixture.staff)

        response = self.client.post(factories.CustomerFactory.get_list_url(), self._get_valid_payload())

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        customer = Customer.objects.get(uuid=response.data['uuid'])
        self.assertEqual(customer.projects.count(), 0)

    @override_waldur_core_settings(OWNER_CAN_MANAGE_CUSTOMER=True)
    def test_user_can_create_customer_if_he_is_not_staff_if_settings_are_tweaked(self):
        self.client.force_authenticate(user=self.fixture.user)
        response = self.client.post(factories.CustomerFactory.get_list_url(), self._get_valid_payload())

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        # User became owner of created customer
        self.assertEqual(response.data['owners'][0]['uuid'], self.fixture.user.uuid.hex)

    def test_user_can_create_customer_if_he_is_staff(self):
        self.client.force_authenticate(user=self.fixture.staff)

        response = self.client.post(factories.CustomerFactory.get_list_url(), self._get_valid_payload())

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    # Mutation tests
    @data('manager', 'admin', 'customer_support', 'project_support', 'global_support')
    def test_user_cannot_change_customer_as_whole(self, user):
        self.client.force_authenticate(user=getattr(self.fixture, user))

        response = self.client.put(self._get_customer_url(self.fixture.customer),
                                   self._get_valid_payload())

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_user_cannot_change_customer_he_is_owner_of(self):
        self.client.force_authenticate(user=self.fixture.owner)

        response = self.client.put(self._get_customer_url(self.fixture.customer),
                                   self._get_valid_payload())

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_user_can_change_customer_as_whole_if_he_is_staff(self):
        self.client.force_authenticate(user=self.fixture.staff)

        response = self.client.put(self._get_customer_url(self.fixture.customer),
                                   self._get_valid_payload())
        self.assertEqual(response.status_code, status.HTTP_200_OK, 'Error message: %s' % response.data)

    def test_user_cannot_change_single_customer_field_he_is_not_owner_of(self):
        self.client.force_authenticate(user=self.fixture.user)

        self._check_single_customer_field_change_permission(self.fixture.customer, status.HTTP_404_NOT_FOUND)

    def test_user_cannot_change_customer_field_he_is_owner_of(self):
        self.client.force_authenticate(user=self.fixture.owner)

        self._check_single_customer_field_change_permission(self.fixture.customer, status.HTTP_403_FORBIDDEN)

    def test_user_can_change_single_customer_field_if_he_is_staff(self):
        self.client.force_authenticate(user=self.fixture.staff)
        self._check_single_customer_field_change_permission(self.fixture.customer, status.HTTP_200_OK)

    # Helper methods
    def _get_valid_payload(self, resource=None):
        resource = resource or factories.CustomerFactory()

        return {
            'name': resource.name,
            'abbreviation': resource.abbreviation,
            'contact_details': resource.contact_details,
        }

    def _check_single_customer_field_change_permission(self, customer, status_code):
        payload = self._get_valid_payload(customer)

        for field, value in payload.items():
            data = {
                field: value
            }

            response = self.client.patch(self._get_customer_url(customer), data)
            self.assertEqual(response.status_code, status_code)


class CustomerQuotasTest(test.APITransactionTestCase):

    def setUp(self):
        self.customer = factories.CustomerFactory()
        self.staff = factories.UserFactory(is_staff=True)

    def test_staff_can_edit_customer_quotas(self):
        self.client.force_login(self.staff)
        quota = self.customer.quotas.get(name=Customer.Quotas.nc_project_count)

        url = quota_factories.QuotaFactory.get_url(quota)
        response = self.client.put(url, {
            'limit': 100
        })
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        quota.refresh_from_db()
        self.assertEqual(quota.limit, 100)

    def test_owner_can_not_edit_customer_quotas(self):
        owner = factories.UserFactory()
        self.customer.add_user(owner, CustomerRole.OWNER)

        self.client.force_login(owner)
        quota = self.customer.quotas.get(name=Customer.Quotas.nc_project_count)

        url = quota_factories.QuotaFactory.get_url(quota)
        response = self.client.put(url, {
            'limit': 100
        })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        quota.refresh_from_db()
        self.assertNotEqual(quota.limit, 100)

    def test_customer_projects_quota_increases_on_project_creation(self):
        factories.ProjectFactory(customer=self.customer)
        self.assert_quota_usage('nc_project_count', 1)

    def test_customer_projects_quota_decreases_on_project_deletion(self):
        project = factories.ProjectFactory(customer=self.customer)
        project.delete()
        self.assert_quota_usage('nc_project_count', 0)

    def test_customer_services_quota_increases_on_service_creation(self):
        factories.TestServiceFactory(customer=self.customer)
        self.assert_quota_usage('nc_service_count', 1)

    def test_customer_services_quota_decreases_on_service_deletion(self):
        service = factories.TestServiceFactory(customer=self.customer)
        service.delete()
        self.assert_quota_usage('nc_service_count', 0)

    def test_customer_and_project_service_project_link_quota_updated(self):
        self.assert_quota_usage('nc_service_project_link_count', 0)
        service = factories.TestServiceFactory(customer=self.customer)

        project1 = factories.ProjectFactory(customer=self.customer)
        factories.TestServiceProjectLinkFactory(service=service, project=project1)

        project2 = factories.ProjectFactory(customer=self.customer)
        factories.TestServiceProjectLinkFactory(service=service, project=project2)

        self.assertEqual(project1.quotas.get(name='nc_service_project_link_count').usage, 1)
        self.assertEqual(project2.quotas.get(name='nc_service_project_link_count').usage, 1)

        self.assert_quota_usage('nc_service_project_link_count', 2)
        self.assert_quota_usage('nc_service_count', 1)

        project2.delete()
        project1.delete()

        self.assert_quota_usage('nc_service_count', 1)
        self.assert_quota_usage('nc_service_project_link_count', 0)

    def test_customer_users_quota_increases_on_adding_owner(self):
        user = factories.UserFactory()
        self.customer.add_user(user, CustomerRole.OWNER)
        self.assert_quota_usage('nc_user_count', 1)

    def test_customer_users_quota_decreases_on_removing_owner(self):
        user = factories.UserFactory()
        self.customer.add_user(user, CustomerRole.OWNER)
        self.customer.remove_user(user)
        self.assert_quota_usage('nc_user_count', 0)

    def test_customer_users_quota_increases_on_adding_administrator(self):
        project = factories.ProjectFactory(customer=self.customer)
        user = factories.UserFactory()
        project.add_user(user, ProjectRole.ADMINISTRATOR)
        self.assert_quota_usage('nc_user_count', 1)

    def test_customer_users_quota_decreases_on_removing_administrator(self):
        project = factories.ProjectFactory(customer=self.customer)
        user = factories.UserFactory()
        project.add_user(user, ProjectRole.ADMINISTRATOR)
        project.remove_user(user)
        self.assert_quota_usage('nc_user_count', 0)

    def test_customer_quota_is_not_increased_on_adding_owner_as_administrator(self):
        user = factories.UserFactory()
        project = factories.ProjectFactory(customer=self.customer)
        self.customer.add_user(user, CustomerRole.OWNER)
        project.add_user(user, ProjectRole.ADMINISTRATOR)

        self.assert_quota_usage('nc_user_count', 1)

    def test_customer_quota_is_not_increased_on_adding_owner_as_manager(self):
        user = factories.UserFactory()
        project = factories.ProjectFactory(customer=self.customer)
        self.customer.add_user(user, CustomerRole.OWNER)
        project.add_user(user, ProjectRole.ADMINISTRATOR)

        self.assert_quota_usage('nc_user_count', 1)

    def test_customer_users_quota_decreases_when_one_project_is_deleted(self):
        project = factories.ProjectFactory(customer=self.customer)
        user = factories.UserFactory()

        project.add_user(user, ProjectRole.ADMINISTRATOR)
        self.assert_quota_usage('nc_user_count', 1)

        project.delete()
        self.assert_quota_usage('nc_user_count', 0)

    def test_customer_users_quota_decreases_when_projects_are_deleted_in_bulk(self):
        count = 2
        for _ in range(count):
            project = factories.ProjectFactory(customer=self.customer)
            user = factories.UserFactory()
            project.add_user(user, ProjectRole.ADMINISTRATOR)

        self.assert_quota_usage('nc_user_count', count)

        self.customer.projects.all().delete()
        self.assert_quota_usage('nc_user_count', 0)

    def assert_quota_usage(self, name, value):
        self.assertEqual(value, self.customer.quotas.get(name=name).usage)


class CustomerUnicodeTest(TransactionTestCase):
    def test_customer_can_have_unicode_name(self):
        factories.CustomerFactory(name="Моя организация")


@ddt
class CustomerUsersListTest(test.APITransactionTestCase):
    all_users = ('staff', 'owner', 'manager', 'admin', 'customer_support', 'project_support', 'global_support')

    def setUp(self):
        self.fixture = fixtures.ProjectFixture()
        self.url = factories.CustomerFactory.get_url(self.fixture.customer, action='users')

    @data(*all_users)
    def test_user_can_list_customer_users(self, user):
        self.client.force_authenticate(getattr(self.fixture, user))
        # call fixture to initiate all users:
        for user in self.all_users:
            getattr(self.fixture, user)
        response = self.client.get(self.url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 5)

        self.assertSetEqual({user['role'] for user in response.data}, {'owner', 'support', None, None})
        self.assertSetEqual({user['uuid'] for user in response.data},
                            {self.fixture.owner.uuid.hex, self.fixture.admin.uuid.hex, self.fixture.manager.uuid.hex,
                             self.fixture.customer_support.uuid.hex, self.fixture.project_support.uuid.hex})
        self.assertSetEqual({user['projects'] and user['projects'][0]['role'] or None
                             for user in response.data}, {None, 'admin', 'manager', 'support'})

    def test_user_can_not_list_project_users(self):
        self.client.force_authenticate(self.fixture.user)
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_users_ordering_by_concatenated_name(self):
        walter = factories.UserFactory(full_name='', username='walter')
        admin = factories.UserFactory(full_name='admin', username='zzz')
        alice = factories.UserFactory(full_name='', username='alice')
        dave = factories.UserFactory(full_name='dave', username='dave')
        expected_order = [admin, alice, dave, walter]
        for user in expected_order:
            self.fixture.customer.add_user(user, CustomerRole.OWNER)

        self.client.force_authenticate(self.fixture.staff)
        response = self.client.get(self.url + '?o=concatenated_name')
        for serialized_user, expected_user in zip(response.data, expected_order):
            self.assertEqual(serialized_user['uuid'], expected_user.uuid.hex)

        # reversed order
        response = self.client.get(self.url + '?o=-concatenated_name')
        for serialized_user, expected_user in zip(response.data, expected_order[::-1]):
            self.assertEqual(serialized_user['uuid'], expected_user.uuid.hex)

    def test_filter_by_email(self):
        walter = factories.UserFactory(full_name='', username='walter', email='walter@gmail.com')
        admin = factories.UserFactory(full_name='admin', username='zzz', email='admin@waldur.com')
        alice = factories.UserFactory(full_name='', username='alice', email='alice@gmail.com')

        for user in [admin, alice, walter]:
            self.fixture.customer.add_user(user, CustomerRole.OWNER)
        self.client.force_authenticate(self.fixture.staff)

        response = self.client.get(self.url, {'email': 'gmail.com'})
        self.assertEqual(len(response.data), 2)


@ddt
class CustomerCountersListTest(test.APITransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ServiceFixture()
        self.owner = self.fixture.owner
        self.customer_support = self.fixture.customer_support
        self.admin = self.fixture.admin
        self.manager = self.fixture.manager
        self.project_support = self.fixture.project_support
        self.customer = self.fixture.customer
        self.service = self.fixture.service
        self.url = factories.CustomerFactory.get_url(self.customer, action='counters')

    @data('owner', 'customer_support')
    def test_user_can_get_customer_counters(self, user):
        self.client.force_authenticate(getattr(self.fixture, user))
        response = self.client.get(self.url, {'fields': ['users', 'projects', 'services']})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'users': 5, 'projects': 1, 'services': 1})


class UserCustomersFilterTest(test.APITransactionTestCase):
    def setUp(self):
        self.staff = factories.UserFactory(is_staff=True)
        self.user1 = factories.UserFactory()
        self.user2 = factories.UserFactory()

        self.customer1 = factories.CustomerFactory()
        self.customer2 = factories.CustomerFactory()

        self.customer1.add_user(self.user1, CustomerRole.OWNER)
        self.customer2.add_user(self.user1, CustomerRole.OWNER)
        self.customer2.add_user(self.user2, CustomerRole.OWNER)

    def test_staff_can_filter_customer_by_user(self):
        self.assert_staff_can_filter_customer_by_user(self.user1, {self.customer1, self.customer2})
        self.assert_staff_can_filter_customer_by_user(self.user2, {self.customer2})

    def assert_staff_can_filter_customer_by_user(self, user, customers):
        self.client.force_authenticate(self.staff)
        response = self.client.get(factories.CustomerFactory.get_list_url(), {
            'user_uuid': user.uuid.hex,
            'fields': ['uuid']
        })
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual({customer['uuid'] for customer in response.data},
                         {customer.uuid.hex for customer in customers})



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import waldur_core.structure.models


def migrate_roles(apps, schema_editor):
    Invitation = apps.get_model('users', 'Invitation')
    CUSTOMER_ROLES = {
        0: 'owner'
    }
    PROJECT_ROLES = {
        0: 'admin',
        1: 'manager'
    }
    for invitation in Invitation.objects.all():
        if invitation.customer_role:
            invitation.new_customer_role = CUSTOMER_ROLES[invitation.customer_role.role_type]
        if invitation.project_role:
            invitation.project = invitation.project_role.project
            invitation.new_project_role = PROJECT_ROLES[invitation.project_role.role_type]
        invitation.save()


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0038_add_project_and_customer_permissions'),
        ('users', '0003_invitation_civil_number'),
    ]

    operations = [
        migrations.AddField(
            model_name='invitation',
            name='created_by',
            field=models.ForeignKey(related_name='+', to=settings.AUTH_USER_MODEL, blank=True, null=True),
        ),
        migrations.AddField(
            model_name='invitation',
            name='project',
            field=models.ForeignKey(related_name='invitations', blank=True, to='structure.Project', null=True),
        ),
        migrations.AddField(
            model_name='invitation',
            name='new_customer_role',
            field=waldur_core.structure.models.CustomerRole(verbose_name='organization role', blank=True, max_length=30, null=True, choices=[('owner', 'Owner')]),
        ),
        migrations.AddField(
            model_name='invitation',
            name='new_project_role',
            field=waldur_core.structure.models.ProjectRole(blank=True, max_length=30, null=True, choices=[('admin', 'Administrator'), ('manager', 'Manager')]),
        ),
        migrations.RunPython(migrate_roles),
        migrations.RemoveField(
            model_name='invitation',
            name='customer_role',
        ),
        migrations.RemoveField(
            model_name='invitation',
            name='project_role',
        ),
        migrations.RenameField(
            model_name='invitation',
            old_name='new_customer_role',
            new_name='customer_role',
        ),
        migrations.RenameField(
            model_name='invitation',
            old_name='new_project_role',
            new_name='project_role',
        ),
    ]



from django.test import TransactionTestCase

from waldur_core.structure.tests import fixtures


class TotalVolumeSizeTest(TransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ServiceFixture()

    def test_total_volume_size_for_project(self):
        volume = self.fixture.volume
        scope = self.fixture.project
        quota = scope.quotas.get(name='nc_volume_size')
        self.assertEqual(quota.usage, volume.size)

    def test_total_volume_size_for_customer(self):
        volume = self.fixture.volume
        scope = self.fixture.customer
        quota = scope.quotas.get(name='nc_volume_size')
        self.assertEqual(quota.usage, volume.size)

    def test_total_snapshot_size_for_project(self):
        snapshot = self.fixture.snapshot
        scope = self.fixture.project
        quota = scope.quotas.get(name='nc_snapshot_size')
        self.assertEqual(quota.usage, snapshot.size)

    def test_total_snapshot_size_for_customer(self):
        snapshot = self.fixture.snapshot
        scope = self.fixture.customer
        quota = scope.quotas.get(name='nc_snapshot_size')
        self.assertEqual(quota.usage, snapshot.size)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0024_add_sugarcrm_to_settings'),
    ]

    operations = [
        migrations.AlterField(
            model_name='servicesettings',
            name='type',
            field=models.SmallIntegerField(choices=[(1, b'OpenStack'), (2, b'DigitalOcean'), (3, b'Amazon'), (4, b'Jira'), (5, b'GitLab'), (6, b'Oracle'), (7, b'Azure'), (8, b'SugarCRM'), (9, b'SaltStack'), (10, b'Zabbix')]),
            preserve_default=True,
        ),
    ]



from django.test import TestCase

from waldur_core.structure import SupportedServices, ServiceBackendNotImplemented
from waldur_core.structure.tests import TestConfig, TestBackend
from waldur_core.structure.tests.models import TestService, TestNewInstance
from waldur_core.structure.tests.serializers import ServiceSerializer


class ServiceRegistryTest(TestCase):
    def test_invalid_service_type(self):
        self.assertRaises(ServiceBackendNotImplemented,
                          SupportedServices.get_service_backend, 'invalid')

    def test_get_service_backend(self):
        self.assertEqual(TestBackend,
                         SupportedServices.get_service_backend(TestConfig.service_name))

    def test_get_service_serializer(self):
        self.assertEqual(ServiceSerializer,
                         SupportedServices.get_service_serializer(TestService))

    def test_get_service_resources(self):
        self.assertEqual([TestNewInstance],
                         SupportedServices.get_service_resources(TestService))

    def test_model_key(self):
        self.assertEqual(TestConfig.service_name,
                         SupportedServices.get_model_key(TestNewInstance))



import logging

from celery import shared_task
from django.conf import settings
from django.utils import timezone

from waldur_core.logging.loggers import alert_logger, event_logger
from waldur_core.logging.models import BaseHook, Alert, AlertThresholdMixin

logger = logging.getLogger(__name__)


@shared_task(name='waldur_core.logging.process_event')
def process_event(event):
    for hook in BaseHook.get_active_hooks():
        if check_event(event, hook):
            hook.process(event)


def check_event(event, hook):
    # Check that event matches with hook
    if event['type'] not in hook.all_event_types:
        return False
    for key, uuids in event_logger.get_permitted_objects_uuids(hook.user).items():
        if key in event['context'] and event['context'][key] in uuids:
            return True
    return False


@shared_task(name='waldur_core.logging.close_alerts_without_scope')
def close_alerts_without_scope():
    for alert in Alert.objects.filter(closed__isnull=True).iterator():
        if alert.scope is None:
            logger.error('Alert without scope was not closed. Alert id: %s.', alert.id)
            alert.close()


@shared_task(name='waldur_core.logging.alerts_cleanup')
def alerts_cleanup():
    timespan = settings.WALDUR_CORE.get('CLOSED_ALERTS_LIFETIME')
    if timespan:
        Alert.objects.filter(closed__lte=timezone.now() - timespan).delete()


@shared_task(name='waldur_core.logging.check_threshold')
def check_threshold():
    for model in AlertThresholdMixin.get_all_models():
        for obj in model.get_checkable_objects().filter(threshold__gt=0).iterator():
            if obj.is_over_threshold() and obj.scope:
                alert_logger.threshold.warning(
                    'Threshold for {scope_name} is exceeded.',
                    scope=obj.scope,
                    alert_type='threshold_exceeded',
                    alert_context={
                        'object': obj
                    })
            else:
                alert_logger.threshold.close(
                    scope=obj.scope,
                    alert_type='threshold_exceeded')



from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.db.models.signals import pre_save, pre_init, pre_delete, post_save, post_delete, post_init


class Command(BaseCommand):
    help = "Load data with disabled signals."

    def add_arguments(self, parser):
        parser.add_argument(
            '--path', '-p', dest='path', default='nc.json', help='Path to dumped database.'
        )

    def handle(self, *args, **options):
        path = options.get('path')
        for signal in [pre_save, pre_init, pre_delete, post_save, post_delete, post_init]:
            signal.receivers = []
        call_command('loaddata', path)



from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import permissions as rf_permissions
from rest_framework import viewsets

from . import serializers, filters, models


class ResourceSlaStateTransitionViewSet(viewsets.ReadOnlyModelViewSet):
    """
    Service SLAs are connected with occurrences of events.
    To get a list of such events issue a **GET** request to */api/resource-sla-state-transition/*.

    The output contains a list of states and timestamps when the state was reached.
    The list is sorted in descending order by the timestamp.
    """
    queryset = models.ResourceSlaStateTransition.objects.all().order_by('-timestamp')
    serializer_class = serializers.ResourceSlaStateTransitionSerializer
    permission_classes = (rf_permissions.IsAuthenticated,)
    filter_backends = (filters.ResourceScopeFilterBackend, DjangoFilterBackend)
    filter_class = filters.ResourceStateFilter



from django.test import TestCase

from .. import factories as structure_factories


class TagCacheTask(TestCase):
    def test_cache_populated_when_tag_added(self):
        settings = structure_factories.ServiceSettingsFactory()
        settings.tags.add('IAAS')
        self.assertEqual(settings.get_tags(), ['IAAS'])

    def test_cache_cleaned_when_tag_removed(self):
        settings = structure_factories.ServiceSettingsFactory()
        settings.tags.add('IAAS')
        settings.tags.remove('IAAS')
        self.assertEqual(settings.get_tags(), [])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('cost_tracking', '0014_more_digits_for_price'),
    ]

    operations = [
        migrations.AddField(
            model_name='defaultpricelistitem',
            name='metadata',
            field=waldur_core.core.fields.JSONField(blank=True, default={}, help_text='Details of the item, that corresponds price list item. Example: details of flavor.'),
            preserve_default=False,
        ),
    ]



from __future__ import unicode_literals

from ddt import data, ddt
from django.test import TransactionTestCase
from django.urls import reverse
from mock_django import mock_signal_receiver
from rest_framework import status, test
from six.moves import mock

from waldur_core.quotas.tests import factories as quota_factories
from waldur_core.structure import executors, models, signals, views
from waldur_core.structure.models import CustomerRole, Project, ProjectRole
from waldur_core.structure.tests import factories, fixtures, models as test_models


class ProjectPermissionGrantTest(TransactionTestCase):
    def setUp(self):
        self.project = factories.ProjectFactory()
        self.user = factories.UserFactory()

    def test_add_user_returns_created_if_grant_didnt_exist_before(self):
        _, created = self.project.add_user(self.user, ProjectRole.ADMINISTRATOR)

        self.assertTrue(created, 'Project permission should have been reported as created')

    def test_add_user_returns_not_created_if_grant_existed_before(self):
        self.project.add_user(self.user, ProjectRole.ADMINISTRATOR)
        _, created = self.project.add_user(self.user, ProjectRole.ADMINISTRATOR)

        self.assertFalse(created, 'Project permission should have been reported as not created')

    def test_add_user_returns_membership(self):
        membership, _ = self.project.add_user(self.user, ProjectRole.ADMINISTRATOR)

        self.assertEqual(membership.user, self.user)
        self.assertEqual(membership.project, self.project)

    def test_add_user_returns_same_membership_for_consequent_calls_with_same_arguments(self):
        membership1, _ = self.project.add_user(self.user, ProjectRole.ADMINISTRATOR)
        membership2, _ = self.project.add_user(self.user, ProjectRole.ADMINISTRATOR)

        self.assertEqual(membership1, membership2)

    def test_add_user_emits_structure_role_granted_if_grant_didnt_exist_before(self):
        with mock_signal_receiver(signals.structure_role_granted) as receiver:
            self.project.add_user(self.user, ProjectRole.ADMINISTRATOR)

        receiver.assert_called_once_with(
            structure=self.project,
            user=self.user,
            role=ProjectRole.ADMINISTRATOR,
            created_by=None,

            sender=Project,
            signal=signals.structure_role_granted,
        )

    def test_add_user_doesnt_emit_structure_role_granted_if_grant_existed_before(self):
        self.project.add_user(self.user, ProjectRole.ADMINISTRATOR)

        with mock_signal_receiver(signals.structure_role_granted) as receiver:
            self.project.add_user(self.user, ProjectRole.ADMINISTRATOR)

        self.assertFalse(receiver.called, 'structure_role_granted should not be emitted')


class ProjectPermissionRevokeTest(TransactionTestCase):
    def setUp(self):
        self.project = factories.ProjectFactory()
        self.user = factories.UserFactory()
        self.removed_by = factories.UserFactory()

    def test_remove_user_emits_structure_role_revoked_for_each_role_user_had_in_project(self):
        self.project.add_user(self.user, ProjectRole.ADMINISTRATOR)
        self.project.add_user(self.user, ProjectRole.MANAGER)

        with mock_signal_receiver(signals.structure_role_revoked) as receiver:
            self.project.remove_user(self.user, removed_by=self.removed_by)

        calls = [
            mock.call(
                structure=self.project,
                user=self.user,
                removed_by=self.removed_by,
                role=ProjectRole.MANAGER,

                sender=Project,
                signal=signals.structure_role_revoked,
            ),

            mock.call(
                structure=self.project,
                user=self.user,
                removed_by=self.removed_by,
                role=ProjectRole.ADMINISTRATOR,

                sender=Project,
                signal=signals.structure_role_revoked,
            ),
        ]

        receiver.assert_has_calls(calls, any_order=True)

        self.assertEqual(
            receiver.call_count, 2,
            'Excepted exactly 2 signals emitted'
        )

    def test_remove_user_emits_structure_role_revoked_if_grant_existed_before(self):
        self.project.add_user(self.user, ProjectRole.MANAGER)

        with mock_signal_receiver(signals.structure_role_revoked) as receiver:
            self.project.remove_user(self.user, ProjectRole.MANAGER, self.removed_by)

        receiver.assert_called_once_with(
            structure=self.project,
            user=self.user,
            role=ProjectRole.MANAGER,

            sender=Project,
            signal=signals.structure_role_revoked,
            removed_by=self.removed_by,
        )

    def test_remove_user_doesnt_emit_structure_role_revoked_if_grant_didnt_exist_before(self):
        with mock_signal_receiver(signals.structure_role_revoked) as receiver:
            self.project.remove_user(self.user, ProjectRole.MANAGER)

        self.assertFalse(receiver.called, 'structure_role_remove should not be emitted')


class ProjectUpdateDeleteTest(test.APITransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ServiceFixture()

    # Update tests:
    def test_user_can_change_single_project_field(self):
        self.client.force_authenticate(self.fixture.staff)

        data = {'name': 'New project name'}
        response = self.client.patch(factories.ProjectFactory.get_url(self.fixture.project), data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('New project name', response.data['name'])
        self.assertTrue(Project.objects.filter(name=data['name']).exists())

    # Delete tests:
    def test_user_can_delete_project_belonging_to_the_customer_he_owns(self):
        self.client.force_authenticate(self.fixture.owner)

        project = self.fixture.project
        response = self.client.delete(factories.ProjectFactory.get_url(project))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(Project.objects.filter(pk=project.pk).exists())


class ProjectCreateTest(test.APITransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ServiceFixture()

    def test_staff_can_create_any_project(self):
        self.client.force_authenticate(self.fixture.owner)
        data = self._get_valid_project_payload(self.fixture.customer)

        response = self.client.post(factories.ProjectFactory.get_list_url(), data)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertTrue(Project.objects.filter(name=data['name']).exists())

    def test_owner_can_create_project_belonging_to_the_customer_he_owns(self):
        self.client.force_authenticate(self.fixture.owner)
        data = self._get_valid_project_payload(self.fixture.customer)

        response = self.client.post(factories.ProjectFactory.get_list_url(), data)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertTrue(Project.objects.filter(name=data['name']).exists())

    def test_owner_cannot_create_project_not_belonging_to_the_customer_he_owns(self):
        self.client.force_authenticate(self.fixture.owner)
        data = self._get_valid_project_payload(factories.CustomerFactory())
        data['name'] = 'unique name 2'

        response = self.client.post(factories.ProjectFactory.get_list_url(), data)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertFalse(Project.objects.filter(name=data['name']).exists())

    def test_owner_cannot_create_project_if_customer_quota_were_exceeded(self):
        self.fixture.customer.set_quota_limit('nc_project_count', 0)
        data = self._get_valid_project_payload(self.fixture.customer)
        self.client.force_authenticate(self.fixture.owner)

        response = self.client.post(factories.ProjectFactory.get_list_url(), data)

        self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)

    def test_customer_support_cannot_create_project(self):
        self.client.force_authenticate(self.fixture.customer_support)
        data = self._get_valid_project_payload(self.fixture.customer)

        response = self.client.post(factories.ProjectFactory.get_list_url(), data)

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(Project.objects.filter(name=data['name']).exists())

    def test_user_can_specify_certifications(self):
        self.client.force_authenticate(self.fixture.owner)
        data = self._get_valid_project_payload(self.fixture.customer)
        certificate = factories.ServiceCertificationFactory()
        data['certifications'] = [{"url": factories.ServiceCertificationFactory.get_url(certificate)}]

        response = self.client.post(factories.ProjectFactory.get_list_url(), data)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertTrue(Project.objects.filter(name=data['name'], customer=self.fixture.customer).exists())
        self.assertTrue(models.ServiceCertification.objects.filter(projects__name=data['name'],
                                                                   name=certificate.name).exists())

    def _get_valid_project_payload(self, customer):
        return {
            'name': 'New project name',
            'customer': factories.CustomerFactory.get_url(customer),
        }


class ProjectApiPermissionTest(test.APITransactionTestCase):
    forbidden_combinations = (
        # User role, Project
        ('admin', 'manager'),
        ('admin', 'inaccessible'),
        ('manager', 'admin'),
        ('manager', 'inaccessible'),
        ('no_role', 'admin'),
        ('no_role', 'manager'),
        ('no_role', 'inaccessible'),
    )

    def setUp(self):
        self.users = {
            'owner': factories.UserFactory(),
            'admin': factories.UserFactory(),
            'manager': factories.UserFactory(),
            'no_role': factories.UserFactory(),
            'multirole': factories.UserFactory(),
        }

        self.projects = {
            'owner': factories.ProjectFactory(),
            'admin': factories.ProjectFactory(),
            'manager': factories.ProjectFactory(),
            'inaccessible': factories.ProjectFactory(),
        }

        self.projects['admin'].add_user(self.users['admin'], ProjectRole.ADMINISTRATOR)
        self.projects['manager'].add_user(self.users['manager'], ProjectRole.MANAGER)

        self.projects['admin'].add_user(self.users['multirole'], ProjectRole.ADMINISTRATOR)
        self.projects['manager'].add_user(self.users['multirole'], ProjectRole.MANAGER)
        self.projects['owner'].customer.add_user(self.users['owner'], CustomerRole.OWNER)

    # TODO: Test for customer owners
    # Creation tests
    def test_anonymous_user_cannot_create_project(self):
        for old_project in self.projects.values():
            project = factories.ProjectFactory(customer=old_project.customer)
            response = self.client.post(reverse('project-list'), self._get_valid_payload(project))
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_user_cannot_create_project_within_customer_he_doesnt_own_but_admins_its_project(self):
        self.client.force_authenticate(user=self.users['admin'])

        customer = self.projects['admin'].customer

        project = factories.ProjectFactory(customer=customer)
        response = self.client.post(reverse('project-list'), self._get_valid_payload(project))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictContainsSubset(
            {'detail': 'You do not have permission to perform this action.'}, response.data)

    def test_user_cannot_create_project_within_customer_he_doesnt_own_but_manages_its_project(self):
        self.client.force_authenticate(user=self.users['manager'])

        customer = self.projects['manager'].customer

        project = factories.ProjectFactory(customer=customer)
        response = self.client.post(reverse('project-list'), self._get_valid_payload(project))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictContainsSubset(
            {'detail': 'You do not have permission to perform this action.'}, response.data)

    def test_user_cannot_create_project_within_customer_he_is_not_affiliated_with(self):
        self.client.force_authenticate(user=self.users['admin'])

        project = factories.ProjectFactory()
        response = self.client.post(reverse('project-list'), self._get_valid_payload(project))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictContainsSubset(
            {'customer': ['Invalid hyperlink - Object does not exist.']}, response.data)

    def test_user_can_create_project_within_customer_he_owns(self):
        self.client.force_authenticate(user=self.users['owner'])

        customer = self.projects['owner'].customer

        project = factories.ProjectFactory(customer=customer)
        response = self.client.post(reverse('project-list'), self._get_valid_payload(project))
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_staff_user_can_create_project(self):
        staff = factories.UserFactory(is_staff=True)
        self.client.force_authenticate(user=staff)

        customer = self.projects['inaccessible'].customer

        project = factories.ProjectFactory(customer=customer)
        response = self.client.post(reverse('project-list'), self._get_valid_payload(project))
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    # List filtration tests
    def test_anonymous_user_cannot_list_projects(self):
        response = self.client.get(reverse('project-list'))
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_user_can_list_projects_belonging_to_customer_he_owns(self):
        self._ensure_list_access_allowed('owner')

    def test_user_can_list_projects_he_is_administrator_of(self):
        self._ensure_list_access_allowed('admin')

    def test_user_can_list_projects_he_is_manager_of(self):
        self._ensure_list_access_allowed('manager')

    def test_user_cannot_list_projects_he_has_no_role_in(self):
        for user_role, project in self.forbidden_combinations:
            self._ensure_list_access_forbidden(user_role, project)

    def test_user_can_filter_by_projects_where_he_has_manager_role(self):
        self.client.force_authenticate(user=self.users['multirole'])
        response = self.client.get(reverse('project-list') + '?can_manage')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        managed_project_url = self._get_project_url(self.projects['manager'])
        administrated_project_url = self._get_project_url(self.projects['admin'])

        self.assertIn(managed_project_url, [resource['url'] for resource in response.data])
        self.assertNotIn(administrated_project_url, [resource['url'] for resource in response.data])

    # Direct instance access tests
    def test_anonymous_user_cannot_access_project(self):
        project = factories.ProjectFactory()
        response = self.client.get(self._get_project_url(project))
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_user_can_access_project_belonging_to_customer_he_owns(self):
        self._ensure_direct_access_allowed('owner')

    def test_user_can_access_project_he_is_administrator_of(self):
        self._ensure_direct_access_allowed('admin')

    def test_user_can_access_project_he_is_manager_of(self):
        self._ensure_direct_access_allowed('manager')

    def test_user_cannot_access_project_he_has_no_role_in(self):
        for user_role, project in self.forbidden_combinations:
            self._ensure_direct_access_forbidden(user_role, project)

    # Helper methods
    def _get_project_url(self, project):
        return factories.ProjectFactory.get_url(project)

    def _get_valid_payload(self, resource=None):
        resource = resource or factories.ProjectFactory()
        return {
            'name': resource.name,
            'customer': factories.CustomerFactory.get_url(resource.customer),
        }

    def _ensure_list_access_allowed(self, user_role):
        self.client.force_authenticate(user=self.users[user_role])

        response = self.client.get(reverse('project-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        project_url = self._get_project_url(self.projects[user_role])
        self.assertIn(project_url, [instance['url'] for instance in response.data])

    def _ensure_list_access_forbidden(self, user_role, project):
        self.client.force_authenticate(user=self.users[user_role])

        response = self.client.get(reverse('project-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        project_url = self._get_project_url(self.projects[project])
        self.assertNotIn(project_url, [resource['url'] for resource in response.data])

    def _ensure_direct_access_allowed(self, user_role):
        self.client.force_authenticate(user=self.users[user_role])
        response = self.client.get(self._get_project_url(self.projects[user_role]))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def _ensure_direct_access_forbidden(self, user_role, project):
        self.client.force_authenticate(user=self.users[user_role])

        response = self.client.get(self._get_project_url(self.projects[project]))
        # 404 is used instead of 403 to hide the fact that the resource exists at all
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)


@ddt
class ProjectUsersListTest(test.APITransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ProjectFixture()
        self.admin = self.fixture.admin
        self.manager = self.fixture.manager
        self.project = self.fixture.project
        self.url = factories.ProjectFactory.get_url(self.project, action='users')

    @data('staff', 'owner', 'manager', 'admin')
    def test_user_can_list_project_users(self, user):
        self.client.force_authenticate(getattr(self.fixture, user))
        response = self.client.get(self.url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 2)

        self.assertSetEqual({user['role'] for user in response.data}, {'admin', 'manager'})
        self.assertSetEqual({user['uuid'] for user in response.data},
                            {self.admin.uuid.hex, self.manager.uuid.hex})

    def test_user_can_not_list_project_users(self):
        self.client.force_authenticate(self.fixture.user)
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)


class ProjectCountersListTest(test.APITransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ServiceFixture()
        self.owner = self.fixture.owner
        self.admin = self.fixture.admin
        self.manager = self.fixture.manager
        self.project = self.fixture.project
        self.service = self.fixture.service
        self.resource = self.fixture.resource
        self.url = factories.ProjectFactory.get_url(self.project, action='counters')

    def test_user_can_get_project_counters(self):
        self.client.force_authenticate(self.fixture.owner)
        response = self.client.get(self.url, {'fields': ['users', 'apps', 'vms']})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'users': 2, 'apps': 0, 'vms': 1})

    def test_additional_counters_could_be_registered(self):
        views.ProjectCountersView.register_counter('test', lambda project: 100)
        self.client.force_authenticate(self.fixture.owner)
        response = self.client.get(self.url, {'fields': ['test']})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'test': 100})


@ddt
class ProjectCertificationUpdateTest(test.APITransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ServiceFixture()
        self.project = self.fixture.project
        self.associated_certification = factories.ServiceCertificationFactory()
        self.project.certifications.add(self.associated_certification)
        self.new_certification = factories.ServiceCertificationFactory()
        self.url = factories.ProjectFactory.get_url(self.project, action='update_certifications')

    @data('staff', 'owner')
    def test_user_can_update_certifications(self, user):
        self.client.force_authenticate(getattr(self.fixture, user))
        payload = self._get_payload(self.new_certification)

        response = self.client.post(self.url, payload)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue(self.project.certifications.filter(pk=self.new_certification.pk).exists())
        self.assertFalse(self.project.certifications.filter(pk=self.associated_certification.pk).exists())

    @data('global_support', 'manager')
    def test_user_cannot_update_certifications_if_he_has_no_permissions(self, user):
        self.client.force_authenticate(getattr(self.fixture, user))
        payload = self._get_payload(self.new_certification)

        response = self.client.post(self.url, payload)

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def _get_payload(self, *certifications):
        urls = [{"url": factories.ServiceCertificationFactory.get_url(c)} for c in certifications]
        return {
            'certifications': urls
        }


class ProjectCertificationGetTest(test.APITransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ServiceFixture()

    def test_service_certification_state_is_ok_if_project_certifications_is_a_subset_of_service_certifications(self):
        self.client.force_authenticate(self.fixture.owner)
        link = self.fixture.service_project_link
        project_certifications = [factories.ServiceCertificationFactory()]
        service_certifications = project_certifications + [factories.ServiceCertificationFactory()]
        link.service.settings.certifications.add(*service_certifications)
        link.project.certifications.add(*project_certifications)
        url = factories.ProjectFactory.get_url(link.project)

        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNotNone('validation_state', response.data['services'][0])
        self.assertEqual(response.data['services'][0]['validation_state'], "OK")

    def test_certification_state_is_erred_if_project_certifications_is_not_a_subset_of_service_certifications(self):
        self.client.force_authenticate(self.fixture.owner)
        link = self.fixture.service_project_link
        service_certification = factories.ServiceCertificationFactory()
        project_certification = factories.ServiceCertificationFactory()
        link.service.settings.certifications.add(service_certification)
        link.project.certifications.add(project_certification)
        url = factories.ProjectFactory.get_url(link.project)

        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNotNone('validation_state', response.data['services'][0])
        self.assertEqual(response.data['services'][0]['validation_state'], "ERRED")

    def test_missing_certification_name_is_in_error_message(self):
        self.client.force_authenticate(self.fixture.owner)
        link = self.fixture.service_project_link
        service_certification = factories.ServiceCertificationFactory()
        project_certification = factories.ServiceCertificationFactory()
        link.service.settings.certifications.add(service_certification)
        link.project.certifications.add(project_certification)
        url = factories.ProjectFactory.get_url(link.project)

        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNotNone('validation_state', response.data['services'][0])
        self.assertEqual(response.data['services'][0]['validation_state'], "ERRED")
        self.assertIn(project_certification.name, response.data['services'][0]['validation_message'])


@ddt
class ProjectQuotasTest(test.APITransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ProjectFixture()
        self.quota = self.fixture.project.quotas.get(name=Project.Quotas.nc_app_count)

    def update_quota(self):
        url = quota_factories.QuotaFactory.get_url(self.quota)
        return self.client.put(url, {
            'limit': 100
        })

    @data('staff', 'owner')
    def test_authorized_user_can_edit_project_quota(self, user):
        self.client.force_login(getattr(self.fixture, user))

        response = self.update_quota()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.quota.refresh_from_db()
        self.assertEqual(self.quota.limit, 100)

    @data('admin', 'manager')
    def test_non_authorized_user_can_not_edit_project_quota(self, user):
        self.client.force_login(getattr(self.fixture, user))

        response = self.update_quota()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.quota.refresh_from_db()
        self.assertNotEqual(self.quota.limit, 100)


class TestExecutor(executors.BaseCleanupExecutor):
    pre_models = (test_models.TestNewInstance,)


@mock.patch('waldur_core.core.WaldurExtension.get_extensions')
class ProjectCleanupTest(test.APITransactionTestCase):

    def test_executors_are_sorted_in_topological_order(self, get_extensions):

        class ParentExecutor(executors.BaseCleanupExecutor):
            pass

        class ParentExtension(object):
            @staticmethod
            def get_cleanup_executor():
                return ParentExecutor

        class ChildExecutor(executors.BaseCleanupExecutor):
            related_executor = ParentExecutor

        class ChildExtension(object):
            @staticmethod
            def get_cleanup_executor():
                return ChildExecutor

        get_extensions.return_value = [ParentExtension, ChildExtension]

        self.assertEqual([ChildExecutor, ParentExecutor],
                         executors.ProjectCleanupExecutor.get_executors())

    def test_project_without_resources_is_deleted(self, get_extensions):
        fixture = fixtures.ServiceFixture()
        project = fixture.project

        get_extensions.return_value = []
        executors.ProjectCleanupExecutor.execute(fixture.project, async=False)

        self.assertFalse(models.Project.objects.filter(id=project.id).exists())

    def test_project_with_resources_without_executors_is_not_deleted(self, get_extensions):
        fixture = fixtures.ServiceFixture()
        project = fixture.project
        resource = fixture.resource

        get_extensions.return_value = []
        executors.ProjectCleanupExecutor.execute(fixture.project, async=False)

        self.assertTrue(models.Project.objects.filter(id=project.id).exists())
        self.assertTrue(test_models.TestNewInstance.objects.filter(id=resource.id).exists())

    def test_project_with_resources_and_executors_is_deleted(self, get_extensions):
        fixture = fixtures.ServiceFixture()
        project = fixture.project
        resource = fixture.resource

        class TestExtension(object):
            @staticmethod
            def get_cleanup_executor():
                return TestExecutor

        get_extensions.return_value = [TestExtension]
        executors.ProjectCleanupExecutor.execute(fixture.project, async=False)

        self.assertFalse(models.Project.objects.filter(id=project.id).exists())
        self.assertFalse(test_models.TestNewInstance.objects.filter(id=resource.id).exists())



from __future__ import unicode_literals

from rest_framework import status
from rest_framework import test

from waldur_core.structure.tests.factories import UserFactory


class VersionApiPermissionTest(test.APITransactionTestCase):
    def setUp(self):
        self.version_url = 'http://testserver/api/version/'

    def test_authenticated_user_can_access_version(self):
        user = UserFactory()
        self.client.force_authenticate(user=user)

        response = self.client.get(self.version_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_anonymous_user_can_access_version(self):
        response = self.client.get(self.version_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)



from __future__ import unicode_literals

import collections
import datetime

from django.contrib.auth import get_user_model
from django.urls import reverse
from django.utils import timezone
from rest_framework import status, test
from six.moves import mock

from waldur_core.structure import tasks
from waldur_core.structure.models import ProjectRole, CustomerRole, ProjectPermission
from waldur_core.structure.tests import factories, fixtures

User = get_user_model()

TestRole = collections.namedtuple('TestRole', ['user', 'project', 'role'])


class ProjectPermissionBaseTest(test.APITransactionTestCase):
    all_roles = (
        #           user      project     role
        TestRole('admin1', 'project11', 'admin'),
        TestRole('admin2', 'project11', 'admin'),
        TestRole('admin3', 'project12', 'admin'),
        TestRole('admin4', 'project13', 'admin'),
        TestRole('admin5', 'project21', 'admin'),
        TestRole('project_manager1', 'project11', 'manager'),
        TestRole('project_manager2', 'project11', 'manager'),
        TestRole('project_manager3', 'project12', 'manager'),
        TestRole('project_manager4', 'project13', 'manager'),
        TestRole('project_manager5', 'project21', 'manager'),
    )

    role_map = {
        'admin': ProjectRole.ADMINISTRATOR,
        'manager': ProjectRole.MANAGER,
    }

    def setUp(self):
        customers = {
            'customer1': factories.CustomerFactory(),
            'customer2': factories.CustomerFactory(),
        }

        self.projects = {
            'project11': factories.ProjectFactory(customer=customers['customer1']),
            'project12': factories.ProjectFactory(customer=customers['customer1']),
            'project13': factories.ProjectFactory(customer=customers['customer1']),
            'project21': factories.ProjectFactory(customer=customers['customer2']),
        }

        self.users = {
            'owner1': factories.UserFactory(),
            'owner2': factories.UserFactory(),
            'manager1': factories.UserFactory(),
            'manager2': factories.UserFactory(),
            'manager3': factories.UserFactory(),
            'admin1': factories.UserFactory(),
            'admin2': factories.UserFactory(),
            'admin3': factories.UserFactory(),
            'admin4': factories.UserFactory(),
            'admin5': factories.UserFactory(),
            'project_manager1': factories.UserFactory(),
            'project_manager2': factories.UserFactory(),
            'project_manager3': factories.UserFactory(),
            'project_manager4': factories.UserFactory(),
            'project_manager5': factories.UserFactory(),
            'no_role': factories.UserFactory(),
            'staff': factories.UserFactory(is_staff=True),
        }

        customers['customer1'].add_user(self.users['owner1'], CustomerRole.OWNER)
        customers['customer2'].add_user(self.users['owner2'], CustomerRole.OWNER)

        for user, project, role in self.all_roles:
            self.projects[project].add_user(self.users[user], role)

    # Helper methods
    def _get_permission_url(self, user, project, role):
        permission = ProjectPermission.objects.get(
            user=self.users[user],
            role=self.role_map[role],
            project=self.projects[project],
        )
        return 'http://testserver' + reverse('project_permission-detail', kwargs={'pk': permission.pk})


class ProjectPermissionListTest(ProjectPermissionBaseTest):
    def test_anonymous_user_cannot_list_project_permissions(self):
        response = self.client.get(reverse('project_permission-list'))
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_user_cannot_list_roles_of_project_he_is_not_affiliated(self):
        for project in self.projects.keys():
            self.assert_user_access_to_permission_list(user='no_role', project=project, should_see=False)

    def test_customer_owner_can_list_roles_of_his_customers_project(self):
        self.assert_user_access_to_permission_list(user='owner1', project='project11', should_see=True)
        self.assert_user_access_to_permission_list(user='owner1', project='project12', should_see=True)
        self.assert_user_access_to_permission_list(user='owner1', project='project13', should_see=True)

    def test_customer_owner_cannot_list_roles_of_another_customers_project(self):
        self.assert_user_access_to_permission_list(user='owner1', project='project21', should_see=False)

    def test_project_admin_can_list_roles_of_his_project(self):
        self.assert_user_access_to_permission_list(user='admin1', project='project11', should_see=True)

    def test_project_admin_cannot_list_roles_of_another_project(self):
        self.assert_user_access_to_permission_list(user='admin2', project='project12', should_see=False)
        self.assert_user_access_to_permission_list(user='admin2', project='project13', should_see=False)
        self.assert_user_access_to_permission_list(user='admin2', project='project21', should_see=False)

    def test_staff_can_list_roles_of_any_project(self):
        for project in self.projects.keys():
            self.assert_user_access_to_permission_list(user='staff', project=project, should_see=True)

    def assert_user_access_to_permission_list(self, user, project, should_see):
        self.client.force_authenticate(user=self.users[user])

        response = self.client.get(reverse('project_permission-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        expected_urls = {
            r: self._get_permission_url(*r)
            for r in self.all_roles
            if r.project == project
        }

        actual_urls = set([role['url'] for role in response.data])

        for role, role_url in expected_urls.items():
            if should_see:
                self.assertIn(
                    role_url, actual_urls,
                    '{0} user does not see privilege '
                    'he is supposed to see: {1}'.format(user, role),
                )
            else:
                self.assertNotIn(
                    role_url, actual_urls,
                    '{0} user sees privilege '
                    'he is not supposed to see: {1}'.format(user, role),
                )


class ProjectPermissionGrantTest(ProjectPermissionBaseTest):
    def test_customer_owner_can_grant_new_role_within_his_customers_project(self):
        self.assert_user_access_to_permission_granting(
            login_user='owner1',
            affected_user='no_role',
            affected_project='project11',
            expected_status=status.HTTP_201_CREATED,
        )

    def test_customer_owner_can_grant_new_role_within_his_customers_project_for_himself(self):
        self.assert_user_access_to_permission_granting(
            login_user='owner1',
            affected_user='owner1',
            affected_project='project11',
            expected_status=status.HTTP_201_CREATED,
        )

    def test_customer_owner_cannot_grant_existing_role_within_his_project(self):
        self.assert_user_access_to_permission_granting(
            login_user='owner1',
            affected_user='admin1',
            affected_project='project11',
            expected_status=status.HTTP_400_BAD_REQUEST,
            expected_payload={
                'non_field_errors': ['The fields project and user must make a unique set.'],
            }
        )

    def test_customer_owner_cannot_grant_role_within_his_project_if_user_already_has_role(self):
        self.assert_user_access_to_permission_granting(
            login_user='owner1',
            affected_user='admin1',
            role='manager',
            affected_project='project11',
            expected_status=status.HTTP_400_BAD_REQUEST,
            expected_payload={
                'non_field_errors': ['The fields project and user must make a unique set.'],
            }
        )

    def test_customer_owner_cannot_grant_role_within_another_customers_project(self):
        self.assert_user_access_to_permission_granting(
            login_user='owner1',
            affected_user='no_role',
            affected_project='project21',
            expected_status=status.HTTP_400_BAD_REQUEST,
            expected_payload={
                'project': ['Invalid hyperlink - Object does not exist.'],
            }
        )

    def test_project_manager_can_grant_new_admin_role_within_his_project(self):
        self.assert_user_access_to_permission_granting(
            login_user='project_manager1',
            affected_user='no_role',
            affected_project='project11',
            expected_status=status.HTTP_201_CREATED,
            role='admin',
        )

    def test_project_manager_cannot_grant_new_manager_role_within_his_project(self):
        self.assert_user_access_to_permission_granting(
            login_user='project_manager1',
            affected_user='no_role',
            affected_project='project11',
            expected_status=status.HTTP_403_FORBIDDEN,
            role='manager',
            expected_payload={
                'detail': 'You do not have permission to perform this action.',
            }
        )

    def test_project_manager_cannot_grant_new_admin_role_within_not_his_project(self):
        self.assert_user_access_to_permission_granting(
            login_user='project_manager1',
            affected_user='no_role',
            affected_project='project12',
            expected_status=status.HTTP_400_BAD_REQUEST,
            role='admin',
        )

    def test_project_admin_cannot_grant_new_role_within_his_project(self):
        self.assert_user_access_to_permission_granting(
            login_user='admin1',
            affected_user='no_role',
            affected_project='project11',
            expected_status=status.HTTP_403_FORBIDDEN,
            expected_payload={
                'detail': 'You do not have permission to perform this action.',
            }
        )

    def test_project_manager_cannot_grant_existing_role_within_his_project(self):
        self.assert_user_access_to_permission_granting(
            login_user='project_manager1',
            affected_user='admin1',
            affected_project='project11',
            expected_status=status.HTTP_400_BAD_REQUEST,
            expected_payload={
                'non_field_errors': ['The fields project and user must make a unique set.'],
            }
        )

    def test_project_manager_cannot_grant_role_to_himself(self):
        self.assert_user_access_to_permission_granting(
            login_user='project_manager1',
            affected_user='project_manager1',
            affected_project='project11',
            expected_status=status.HTTP_400_BAD_REQUEST,
        )

    def test_project_admin_cannot_grant_role_within_another_project(self):
        self.assert_user_access_to_permission_granting(
            login_user='admin1',
            affected_user='no_role',
            affected_project='project13',
            expected_status=status.HTTP_400_BAD_REQUEST,
            expected_payload={
                'project': ['Invalid hyperlink - Object does not exist.'],
            }
        )

    def test_staff_can_grant_new_role_within_any_project(self):
        for project in self.projects.keys():
            self.assert_user_access_to_permission_granting(
                login_user='staff',
                affected_user='no_role',
                affected_project=project,
                expected_status=status.HTTP_201_CREATED,
            )

    def test_staff_cannot_grant_new_role_if_customer_quota_were_exceeded(self):
        project = 'project11'
        self.projects[project].customer.set_quota_limit('nc_user_count', 0)
        self.assert_user_access_to_permission_granting(
            login_user='staff',
            affected_user='no_role',
            affected_project=project,
            expected_status=status.HTTP_409_CONFLICT,
        )

    def test_staff_cannot_grant_existing_role_within_any_project(self):
        for user, project, role in self.all_roles:
            self.assert_user_access_to_permission_granting(
                login_user='staff',
                affected_user=user,
                affected_project=project,
                expected_status=status.HTTP_400_BAD_REQUEST,
                role=role,
                expected_payload={
                    'non_field_errors': ['The fields project and user must make a unique set.'],
                }
            )

    def assert_user_access_to_permission_granting(self, login_user, affected_user, affected_project,
                                                  expected_status, expected_payload=None, role='admin'):
        self.client.force_authenticate(user=self.users[login_user])

        data = {
            'project': factories.ProjectFactory.get_url(self.projects[affected_project]),
            'user': factories.UserFactory.get_url(self.users[affected_user]),
            'role': role,
        }

        response = self.client.post(reverse('project_permission-list'), data)
        self.assertEqual(response.status_code, expected_status, response.data)
        if expected_payload is not None:
            self.assertDictContainsSubset(expected_payload, response.data)


class ProjectPermissionRevokeTest(ProjectPermissionBaseTest):
    def test_customer_owner_can_revoke_role_within_his_customers_project(self):
        self.assert_user_access_to_permission_revocation(
            login_user='owner1',
            affected_user='admin1',
            affected_project='project11',
            expected_status=status.HTTP_204_NO_CONTENT,
        )

    def test_customer_owner_can_revoke_role_within_his_customers_project_for_himself(self):
        self.projects['project11'].add_user(self.users['owner1'], ProjectRole.ADMINISTRATOR)
        self.assert_user_access_to_permission_revocation(
            login_user='owner1',
            affected_user='owner1',
            affected_project='project11',
            expected_status=status.HTTP_204_NO_CONTENT,
        )

    def test_customer_owner_cannot_revoke_role_within_another_customers_project(self):
        self.assert_user_access_to_permission_revocation(
            login_user='owner1',
            affected_user='admin5',
            affected_project='project21',
            expected_status=status.HTTP_404_NOT_FOUND,
        )

    def test_project_manager_can_revoke_admin_role_within_his_project(self):
        self.assert_user_access_to_permission_revocation(
            login_user='project_manager1',
            affected_user='admin1',
            affected_project='project11',
            expected_status=status.HTTP_204_NO_CONTENT,
        )

    def test_project_manager_cannot_revoke_manager_role_within_his_project(self):
        self.assert_user_access_to_permission_revocation(
            login_user='project_manager1',
            affected_user='project_manager1',
            affected_project='project11',
            expected_status=status.HTTP_403_FORBIDDEN,
            role='manager',
            expected_payload={
                'detail': 'You do not have permission to perform this action.',
            }
        )

    def test_project_admin_cannot_revoke_role_within_his_project(self):
        self.assert_user_access_to_permission_revocation(
            login_user='admin1',
            affected_user='admin2',
            affected_project='project11',
            expected_status=status.HTTP_403_FORBIDDEN,
            expected_payload={
                'detail': 'You do not have permission to perform this action.',
            }
        )

    def test_project_admin_cannot_revoke_role_within_within_another_project(self):
        for user, project, _ in self.all_roles:
            if project == 'project11':
                continue

            self.assert_user_access_to_permission_revocation(
                login_user='admin1',
                affected_user='admin5',
                affected_project='project21',
                expected_status=status.HTTP_404_NOT_FOUND,
            )

    def test_staff_can_revoke_role_within_any_project(self):
        for user, project, role in self.all_roles:
            self.assert_user_access_to_permission_revocation(
                login_user='staff',
                affected_user=user,
                affected_project=project,
                expected_status=status.HTTP_204_NO_CONTENT,
                role=role,
            )

    def assert_user_access_to_permission_revocation(self, login_user, affected_user, affected_project,
                                                    expected_status, expected_payload=None, role='admin'):
        self.client.force_authenticate(user=self.users[login_user])

        url = self._get_permission_url(affected_user, affected_project, role)

        response = self.client.delete(url)
        self.assertEqual(response.status_code, expected_status)
        if expected_payload is not None:
            self.assertDictContainsSubset(expected_payload, response.data)


class ProjectPermissionFilterTest(test.APITransactionTestCase):

    def setUp(self):
        fixture = fixtures.ProjectFixture()
        self.staff = fixture.staff
        self.customer = fixture.customer
        self.admin = fixture.admin
        self.manager = fixture.manager
        self.project = fixture.project
        self.url = reverse('project_permission-list')

    def test_user_can_filter_permission_by_project(self):
        self.client.force_authenticate(self.manager)
        response = self.client.get(self.url, {'project': self.project.uuid.hex})
        self.assertEqual(len(response.data), 2)

    def test_user_can_filter_permission_by_user(self):
        self.client.force_authenticate(self.manager)
        response = self.client.get(self.url, {'user': self.admin.uuid.hex})
        self.assertEqual(len(response.data), 1)

    def test_user_can_filter_permission_by_customer(self):
        self.client.force_authenticate(self.staff)
        response = self.client.get(self.url, {'customer': self.customer.uuid.hex})
        self.assertEqual(len(response.data), 2)

    def test_user_can_filter_permission_by_empty_customer(self):
        self.client.force_authenticate(self.staff)
        response = self.client.get(self.url, {'customer': factories.CustomerFactory().uuid.hex})
        self.assertEqual(len(response.data), 0)


class ProjectPermissionExpirationTest(test.APITransactionTestCase):
    def setUp(self):
        permission = factories.ProjectPermissionFactory()
        self.user = permission.user
        self.project = permission.project
        self.url = reverse('project_permission-detail', kwargs={'pk': permission.pk})

    def test_user_can_not_update_permission_expiration_time(self):
        self.client.force_authenticate(user=self.user)

        expiration_time = timezone.now() + datetime.timedelta(days=100)
        response = self.client.put(self.url, {
            'expiration_time': expiration_time
        })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_staff_can_update_permission_expiration_time(self):
        staff_user = factories.UserFactory(is_staff=True)
        self.client.force_authenticate(user=staff_user)

        expiration_time = timezone.now() + datetime.timedelta(days=100)
        response = self.client.put(self.url, {
            'expiration_time': expiration_time
        })
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['expiration_time'], expiration_time, response.data)

    def test_owner_can_update_permission_for_himself(self):
        owner = factories.UserFactory()
        self.project.customer.add_user(owner, CustomerRole.OWNER)
        self.client.force_authenticate(user=owner)

        expiration_time = timezone.now() + datetime.timedelta(days=100)
        response = self.client.put(self.url, {
            'expiration_time': expiration_time
        })
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['expiration_time'], expiration_time, response.data)

    def test_user_can_set_permission_expiration_time_lower_than_current(self):
        staff_user = factories.UserFactory(is_staff=True)
        self.client.force_authenticate(user=staff_user)

        expiration_time = timezone.now() - datetime.timedelta(days=100)
        response = self.client.put(self.url, {
            'expiration_time': expiration_time
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_user_can_set_expiration_time_role_when_role_is_created(self):
        staff_user = factories.UserFactory(is_staff=True)
        self.client.force_authenticate(user=staff_user)

        expiration_time = timezone.now() + datetime.timedelta(days=100)
        response = self.client.post(factories.ProjectPermissionFactory.get_list_url(), {
            'project': factories.ProjectFactory.get_url(),
            'user': factories.UserFactory.get_url(),
            'role': factories.ProjectPermissionFactory.role,
            'expiration_time': expiration_time,
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data['expiration_time'], expiration_time, response.data)

    def test_user_cannot_grant_permissions_with_greater_expiration_time(self):
        expiration_time = timezone.now() + datetime.timedelta(days=100)
        permission = factories.ProjectPermissionFactory(
            role=ProjectRole.MANAGER,
            expiration_time=expiration_time)
        self.client.force_authenticate(user=permission.user)
        response = self.client.post(factories.ProjectPermissionFactory.get_list_url(), {
            'project': factories.ProjectFactory.get_url(project=permission.project),
            'user': factories.UserFactory.get_url(),
            'role': factories.ProjectPermissionFactory.role,
            'expiration_time': expiration_time + datetime.timedelta(days=1),
        })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_task_revokes_expired_permissions(self):
        expired_permission = factories.ProjectPermissionFactory(
            expiration_time=timezone.now() - datetime.timedelta(days=100))
        not_expired_permission = factories.ProjectPermissionFactory(
            expiration_time=timezone.now() + datetime.timedelta(days=100))
        tasks.check_expired_permissions()

        self.assertFalse(expired_permission.project.has_user(
            expired_permission.user, expired_permission.role))
        self.assertTrue(not_expired_permission.project.has_user(
            not_expired_permission.user, not_expired_permission.role))

    def test_when_expiration_time_is_updated_event_is_emitted(self):
        staff_user = factories.UserFactory(is_staff=True)
        self.client.force_authenticate(user=staff_user)
        expiration_time = timezone.now() + datetime.timedelta(days=100)

        with mock.patch('logging.LoggerAdapter.info') as mocked_info:
            self.client.put(self.url, {'expiration_time': expiration_time})
            (args, kwargs) = mocked_info.call_args_list[-1]
            event_type = kwargs['extra']['event_type']
            event_message = args[0]
            self.assertEqual(event_type, 'role_updated')
            self.assertTrue(staff_user.full_name in event_message)


class ProjectPermissionCreatedByTest(test.APITransactionTestCase):
    def test_user_which_granted_permission_is_stored(self):
        staff_user = factories.UserFactory(is_staff=True)
        self.client.force_authenticate(user=staff_user)

        user = factories.UserFactory()
        project = factories.ProjectFactory()

        data = {
            'project': factories.ProjectFactory.get_url(project),
            'user': factories.UserFactory.get_url(user),
            'role': ProjectRole.ADMINISTRATOR,
        }

        response = self.client.post(reverse('project_permission-list'), data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        permission = ProjectPermission.objects.get(pk=response.data['pk'])
        self.assertEqual(permission.created_by, staff_user)


class GetProjectUsersTest(test.APITransactionTestCase):

    def setUp(self):
        fixture = fixtures.ProjectFixture()
        self.project = fixture.project
        self.admin = fixture.admin
        self.manager = fixture.manager

    def test_get_users_by_default_returns_both_managers_and_admins(self):
        users = list(self.project.get_users())
        self.assertListEqual(users, [self.admin, self.manager])

    def test_get_users_by_returns_admins(self):
        users = list(self.project.get_users(ProjectRole.ADMINISTRATOR))
        self.assertListEqual(users, [self.admin])

    def test_get_users_by_returns_managers(self):
        users = list(self.project.get_users(ProjectRole.MANAGER))
        self.assertListEqual(users, [self.manager])
